from collections.abc import MutableMapping
from collections import namedtuple
import os
import struct

from borghash import HashTableNT

from .constants import UNKNOWN_INT32, UNKNOWN_BYTES32



cdef _NoDefault = object()


class HTProxyMixin:
    def __setitem__(self, key, value):
        self.ht[key] = value

    def __getitem__(self, key):
        return self.ht[key]

    def __delitem__(self, key):
        del self.ht[key]

    def __contains__(self, key):
        return key in self.ht

    def __len__(self):
        return len(self.ht)

    def __iter__(self):
        for key, value in self.ht.items():
            yield key

    def clear(self):
        self.ht.clear()


ChunkIndexEntry = namedtuple('ChunkIndexEntry', 'flags size pack_id obj_offset obj_size')
ChunkIndexEntryFormatT = namedtuple('ChunkIndexEntryFormatT', 'flags size pack_id obj_offset obj_size')
ChunkIndexEntryFormat = ChunkIndexEntryFormatT(flags="I", size="I", pack_id="32s", obj_offset="I", obj_size="I")


class ChunkIndex(HTProxyMixin, MutableMapping):
    """
    Mapping from key256 to (flags32, size32, pack_id256, obj_offset32, obj_size32) to track chunks in the repository.
    """
    # .flags related values:
    F_NONE = 0  # all flags cleared
    M_USER = 0x00ffffff  # mask for user flags
    M_SYSTEM = 0xff000000  # mask for system flags
    # user flags:
    F_USED = 2 ** 0  # chunk is used/referenced
    F_COMPRESS = 2 ** 1  # chunk shall get (re-)compressed
    F_PENDING = 2 ** 2  # pack location (pack_id, obj_offset, obj_size) not resolved yet.
    # system flags (internal use, always 0 to user, not changeable by user):
    F_NEW = 2 ** 24  # a new chunk that is not present in repo index/* yet.

    def __init__(self, capacity=1000, path=None, usable=None):
        if path:
            self.ht = HashTableNT.read(path)
        else:
            if usable is not None:
                capacity = usable * 2  # load factor 0.5
            self.ht = HashTableNT(key_size=32, value_type=ChunkIndexEntry, value_format=ChunkIndexEntryFormat,
                                  capacity=capacity)

    def hide_system_flags(self, value):
        user_flags = value.flags & self.M_USER
        return value._replace(flags=user_flags)

    def iteritems(self, *, only_new=False):
        """Iterates items (optionally only new items); hides system flags."""
        for key, value in self.ht.items():
            if not only_new or (value.flags & self.F_NEW):
                yield key, self.hide_system_flags(value)

    def add(self, key, size):
        v = self.get(key)
        if v is None:
            flags = self.F_USED
        else:
            flags = v.flags | self.F_USED
            assert v.size == 0 or v.size == size
        # F_PENDING marks the pack location (pack_id, obj_offset, obj_size) as not yet set.
        # Re-adding a chunk resets it to UNKNOWN/pending, dropping any prior location until the next flush().
        self[key] = ChunkIndexEntry(
            flags=flags | self.F_PENDING, size=size,
            pack_id=UNKNOWN_BYTES32, obj_offset=UNKNOWN_INT32, obj_size=UNKNOWN_INT32
        )

    def __getitem__(self, key):
        """Specialized __getitem__ that hides system flags."""
        value = self.ht[key]
        return self.hide_system_flags(value)

    def __setitem__(self, key, value):
        """Specialized __setitem__ that protects system flags and manages the F_NEW flag."""
        try:
            prev = self.ht[key]
        except KeyError:
            prev_flags = self.F_NONE
            is_new = True
        else:
            prev_flags = prev.flags
            is_new = bool(prev_flags & self.F_NEW)  # was new? stays new!
        system_flags = prev_flags & self.M_SYSTEM
        if is_new:
            system_flags |= self.F_NEW
        else:
            system_flags &= ~self.F_NEW
        user_flags = value.flags & self.M_USER
        self.ht[key] = value._replace(flags=system_flags | user_flags)

    def update_pack_info(self, pack_results):
        """Set pack_id, obj_offset and obj_size from a list of (chunk_id, pack_id, obj_offset, obj_size)
        tuples and clear F_PENDING."""
        if not pack_results:
            return
        for chunk_id, pack_id, obj_offset, obj_size in pack_results:
            existing = self[chunk_id]
            self[chunk_id] = existing._replace(
                flags=existing.flags & ~self.F_PENDING, pack_id=pack_id, obj_offset=obj_offset, obj_size=obj_size
            )

    def is_pending(self, key):
        """Return whether the chunk's pack location (pack_id, obj_offset, obj_size) is unresolved."""
        return bool(self[key].flags & self.F_PENDING)

    def clear_new(self):
        """Clears the F_NEW flag of all items."""
        for key, value in self.ht.items():
            if value.flags & self.F_NEW:
                flags = value.flags & ~self.F_NEW
                self.ht[key] = value._replace(flags=flags)

    @classmethod
    def read(cls, path):
        return cls(path=path)

    def write(self, path):
        self.ht.write(path)

    def size(self):
        return self.ht.size()

    @property
    def stats(self):
        return self.ht.stats

    def k_to_idx(self, key):
        return self.ht.k_to_idx(key)

    def idx_to_k(self, idx):
        return self.ht.idx_to_k(idx)


FuseVersionsIndexEntry = namedtuple('FuseVersionsIndexEntry', 'version hash')
FuseVersionsIndexEntryFormatT = namedtuple('FuseVersionsIndexEntryFormatT', 'version hash')
FuseVersionsIndexEntryFormat = FuseVersionsIndexEntryFormatT(version="I", hash="16s")


class FuseVersionsIndex(HTProxyMixin, MutableMapping):
    """
    Mapping from key128 to (file_version32, file_content_hash128) to support the FUSE versions view.
    """
    def __init__(self):
        self.ht = HashTableNT(key_size=16, value_type=FuseVersionsIndexEntry, value_format=FuseVersionsIndexEntryFormat)



"""
borg.compress
=============

Compression is applied to chunks after ID hashing (so the ID is a direct function of the
plain chunk, compression is irrelevant to it), and of course before encryption.

The "auto" mode (e.g. --compression auto,lzma,4) is implemented as a meta Compressor,
meaning that Auto acts like a Compressor, but defers actual work to others (namely
LZ4 as a heuristic whether compression is worth it, and the specified Compressor
for the actual compression).

Decompression is normally handled through Compressor.decompress which will detect
which compressor has been used to compress the data and dispatch to the correct
decompressor.
"""

import math
import random
from struct import Struct
import sys
import zlib

try:
    import lzma
except ImportError:
    lzma = None

from .constants import MAX_DATA_SIZE, ROBJ_FILE_STREAM
from .helpers import Buffer, DecompressionError
from .helpers.argparsing import ArgumentTypeError

if sys.version_info >= (3, 14):
    from compression import zstd
else:
    from backports import zstd

cdef extern from "lz4.h":
    int LZ4_compress_default(const char* source, char* dest, int inputSize, int maxOutputSize) nogil
    int LZ4_decompress_safe(const char* source, char* dest, int inputSize, int maxOutputSize) nogil
    int LZ4_compressBound(int inputSize) nogil

buffer = Buffer(bytearray, size=0)

cdef class CompressorBase:
    """
    base class for all (de)compression classes,
    also handles compression format auto detection and
    adding/stripping the ID header (which enables auto-detection).
    """
    ID = 0xFF  # reserved and not used
               # overwrite with a unique 1-byte bytestring in child classes
    name = 'baseclass'

    @classmethod
    def detect(cls, data):
        return data and data[0] == cls.ID

    def __init__(self, level=255, legacy_mode=False, **kwargs):
        assert 0 <= level <= 255
        self.level = level
        self.legacy_mode = legacy_mode  # True: support prefixed ctype/clevel bytes

    def decide(self, data):
        """
        Return which compressor will perform the actual compression for *data*.

        This exists for a very specific case: If borg recreate is instructed to recompress
        using Auto compression it needs to determine the _actual_ target compression of a chunk
        in order to detect whether it should be recompressed.

        Any compressor may return a compressor other than *self*, like e.g. the CNONE compressor,
        and should actually do so if *data* would become larger on compression.
        """
        return self

    def compress(self, meta, data):
        """
        Compress *data* (bytes) and return compression metadata and compressed bytes.
        """
        if not isinstance(data, bytes):
            data = bytes(data)  # code below does not work with memoryview
        if self.legacy_mode:
            return None, bytes((self.ID, self.level)) + data
        else:
            meta["ctype"] = self.ID
            meta["clevel"] = self.level
            meta["csize"] = len(data)
            return meta, data

    def decompress(self, meta, data):
        """
        Decompress *data* (preferably a memoryview, bytes also acceptable) and return bytes result.

        Legacy mode: The leading Compressor ID bytes need to be present.

        Only handles input generated by _this_ Compressor - for a general purpose
        decompression method see *Compressor.decompress*.
        """
        if self.legacy_mode:
            assert meta is None
            meta = {}
            meta["ctype"] = data[0]
            meta["clevel"] = data[1]
            meta["csize"] = len(data)
            return meta, data[2:]
        else:
            assert isinstance(meta, dict)
            assert "ctype" in meta
            assert "clevel" in meta
            return meta, data

    def check_fix_size(self, meta, data):
        if "size" in meta:
            assert meta["size"] == len(data)
        elif self.legacy_mode:
            meta["size"] = len(data)
        else:
            pass  # raise ValueError("size not present and not in legacy mode")

cdef class DecidingCompressor(CompressorBase):
    """
    base class for (de)compression classes that (based on an internal _decide
    method) decide whether and how to compress data.
    """
    name = 'decidebaseclass'

    def __init__(self, level=255, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)

    def _decide(self, meta, data):
        """
        Decides what to do with *data*. Returns (compressor, meta, compressed_data).

        *compressed_data* can be the result of *data* being processed by *compressor*,
        if that is generated as a side-effect of the decision process, or None otherwise.

        This private method allows for more efficient implementation of compress()
        and decide_compress() making use of *compressed_data*, if already generated.
        """
        raise NotImplementedError

    def decide(self, meta, data):
        return self._decide(meta, data)[0]

    def decide_compress(self, meta, data):
        """
        Decides what to do with *data* and handle accordingly. Returns (compressor, compressed_data).

        *compressed_data* is the result of *data* being processed by *compressor*.
        """
        compressor, (meta, compressed_data) = self._decide(meta, data)

        if compressed_data is None:
            meta, compressed_data = compressor.compress(meta, data)

        if compressor is self:
            # call super class to add ID bytes
            return self, super().compress(meta, compressed_data)

        return compressor, (meta, compressed_data)

    def compress(self, meta, data):
        meta["size"] = len(data)
        return self.decide_compress(meta, data)[1]

class CNONE(CompressorBase):
    """
    none - no compression, just pass through data
    """
    ID = 0x00
    name = 'none'

    def __init__(self, level=255, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)  # no defined levels for CNONE, so just say "unknown"

    def compress(self, meta, data):
        meta["size"] = len(data)
        return super().compress(meta, data)

    def decompress(self, meta, data):
        meta, data = super().decompress(meta, data)
        if not isinstance(data, bytes):
            data = bytes(data)
        self.check_fix_size(meta, data)
        return meta, data

class LZ4(DecidingCompressor):
    """
    raw LZ4 compression / decompression (liblz4).

    Features:
        - lz4 is super fast
        - wrapper releases CPython's GIL to support multithreaded code
        - uses safe lz4 methods that never go beyond the end of the output buffer
    """
    ID = 0x01
    name = 'lz4'

    def __init__(self, level=255, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)  # no defined levels for LZ4, so just say "unknown"

    def _decide(self, meta, idata):
        """
        Decides what to do with *data*. Returns (compressor, lz4_data).

        *lz4_data* is the LZ4 result if *compressor* is LZ4 as well, otherwise it is None.
        """
        if not isinstance(idata, bytes):
            idata = bytes(idata)  # code below does not work with memoryview
        cdef int isize = len(idata)
        cdef int osize
        cdef char *source = idata
        cdef char *dest
        osize = LZ4_compressBound(isize)
        buf = buffer.get(osize)
        dest = <char *> buf
        with nogil:
            osize = LZ4_compress_default(source, dest, isize, osize)
        if not osize:
            raise Exception('lz4 compress failed')
        # only compress if the result actually is smaller
        if osize < isize:
            return self, (meta, dest[:osize])
        else:
            return NONE_COMPRESSOR, (meta, None)

    def decompress(self, meta, data):
        meta, idata = super().decompress(meta, data)
        if not isinstance(idata, bytes):
            idata = bytes(idata)  # code below does not work with memoryview
        cdef int isize = len(idata)
        cdef int osize
        cdef int rsize
        cdef char *source = idata
        cdef char *dest
        # a bit more than 8MB is enough for the usual data sizes yielded by the chunker.
        # allocate more if isize * 3 is already bigger, to avoid having to resize often.
        osize = max(int(1.1 * 2**23), isize * 3)
        while True:
            try:
                buf = buffer.get(osize)
            except MemoryError:
                raise DecompressionError('MemoryError')
            dest = <char *> buf
            with nogil:
                rsize = LZ4_decompress_safe(source, dest, isize, osize)
            if rsize >= 0:
                break
            if osize > 2 ** 27:  # 128MiB (should be enough, considering max. repo obj size and very good compression)
                # this is insane, get out of here
                raise DecompressionError('lz4 decompress failed')
            # likely the buffer was too small, get a bigger one:
            osize = int(1.5 * osize)
        data = dest[:rsize]
        self.check_fix_size(meta, data)
        return meta, data

class LZMA(DecidingCompressor):
    """
    lzma compression / decompression
    """
    ID = 0x02
    name = 'lzma'

    def __init__(self, level=6, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)
        self.level = level
        if lzma is None:
            raise ValueError('No lzma support found.')

    def _decide(self, meta, data):
        """
        Decides what to do with *data*. Returns (compressor, lzma_data).

        *lzma_data* is the LZMA result if *compressor* is LZMA as well, otherwise it is None.
        """
        # we do not need integrity checks in lzma, we do that already
        lzma_data = lzma.compress(data, preset=self.level, check=lzma.CHECK_NONE)
        if len(lzma_data) < len(data):
            return self, (meta, lzma_data)
        else:
            return NONE_COMPRESSOR, (meta, None)

    def decompress(self, meta, data):
        meta, data = super().decompress(meta, data)
        try:
            data = lzma.decompress(data)
            self.check_fix_size(meta, data)
            return meta, data
        except lzma.LZMAError as e:
            raise DecompressionError(str(e)) from None

class ZSTD(DecidingCompressor):
    """
    zstd compression / decompression (python stdlib (python >= 3.14))
    """
    ID = 0x03
    name = 'zstd'

    def __init__(self, level=3, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)
        self.level = level

    def _decide(self, meta, data):
        zstd_data = zstd.compress(data, self.level)
        if len(zstd_data) < len(data):
            return self, (meta, zstd_data)
        else:
            return NONE_COMPRESSOR, (meta, None)
    
    def decompress(self, meta, data):
        meta, data = super().decompress(meta, data)
        try:
            data = zstd.decompress(data)
            self.check_fix_size(meta, data)
            return meta, data
        except zstd.ZstdError as e:
            raise DecompressionError(str(e)) from None

class ZLIB(DecidingCompressor):
    """
    zlib compression / decompression (python stdlib)
    """
    ID = 0x05
    name = 'zlib'

    def __init__(self, level=6, legacy_mode=False, **kwargs):
        super().__init__(level=level, legacy_mode=legacy_mode, **kwargs)
        self.level = level

    def _decide(self, meta, data):
        """
        Decides what to do with *data*. Returns (compressor, zlib_data).

        *zlib_data* is the ZLIB result if *compressor* is ZLIB as well, otherwise it is None.
        """
        zlib_data = zlib.compress(data, self.level)
        if len(zlib_data) < len(data):
            return self, (meta, zlib_data)
        else:
            return NONE_COMPRESSOR, (meta, None)

    def decompress(self, meta, data):
        meta, data = super().decompress(meta, data)
        try:
            data = zlib.decompress(data)
            self.check_fix_size(meta, data)
            return meta, data
        except zlib.error as e:
            raise DecompressionError(str(e)) from None

class ZLIB_legacy(CompressorBase):
    """
    zlib compression / decompression (python stdlib)

    Note: This is the legacy ZLIB support as used by borg < 1.3.
          It still suffers from attic *only* supporting zlib and not having separate
          ID bytes to differentiate between differently compressed chunks.
          This just works because zlib compressed stuff always starts with 0x.8.. bytes.
          Newer borg uses the ZLIB class that has separate ID bytes (as all the other
          compressors) and does not need this hack.
    """
    ID = 0x08  # not used here, see detect()
    # avoid all 0x.8 IDs elsewhere!
    name = 'zlib_legacy'

    @classmethod
    def detect(cls, data):
        # matches misc. patterns 0x.8.. used by zlib
        cmf, flg = data[:2]
        is_deflate = cmf & 0x0f == 8
        check_ok = (cmf * 256 + flg) % 31 == 0
        return check_ok and is_deflate

    def __init__(self, level=6, **kwargs):
        super().__init__(level=level, **kwargs)
        self.level = level

    def compress(self, meta, data):
        # note: for compatibility no super call, do not add ID bytes
        return None, zlib.compress(data, self.level)

    def decompress(self, meta, data):
        # note: for compatibility no super call, do not strip ID bytes
        assert self.legacy_mode  # only borg 1.x repos have the legacy ZLIB format
        assert meta is None
        meta = {}
        meta["ctype"] = ZLIB.ID  # change to non-legacy ZLIB id
        meta["clevel"] = 255  # we do not know the compression level
        meta["csize"] = len(data)
        try:
            data = zlib.decompress(data)
            self.check_fix_size(meta, data)
            return meta, data
        except zlib.error as e:
            raise DecompressionError(str(e)) from None

class Auto(CompressorBase):
    """
    Meta-Compressor that decides which compression to use based on LZ4's ratio.

    As a meta-Compressor the actual compression is deferred to other Compressors,
    therefore this Compressor has no ID, no detect() and no decompress().
    """

    ID = None
    name = 'auto'

    def __init__(self, compressor):
        super().__init__()
        self.compressor = compressor

    def _decide(self, meta, data):
        """
        Decides what to do with *data*. Returns (compressor, compressed_data).

        *compressor* is the compressor that is decided to be best suited to compress
        *data*, *compressed_data* is the result of *data* being compressed by a
        compressor, which may or may not be *compressor*!

        There are three possible outcomes of the decision process:
        * *data* compresses well enough for trying the more expensive compressor
          set on instantiation to make sense.
          In this case, (expensive_compressor_class, lz4_compressed_data) is
          returned.
        * *data* compresses only slightly using the LZ4 compressor, thus trying
          the more expensive compressor for potentially little gain does not
          make sense.
          In this case, (LZ4_COMPRESSOR, lz4_compressed_data) is returned.
        * *data* does not compress at all using LZ4, in this case
          (NONE_COMPRESSOR, none_compressed_data) is returned.

        Note: While it makes no sense, the expensive compressor may well be set
        to the LZ4 compressor.
        """
        compressor, (meta, compressed_data) = LZ4_COMPRESSOR.decide_compress(meta, data)
        # compressed_data includes the compression type header, while data does not yet
        ratio = len(compressed_data) / (len(data) + 2)
        if ratio < 0.97:
            return self.compressor, (meta, compressed_data)
        else:
            return compressor, (meta, compressed_data)

    def decide(self, meta, data):
        return self._decide(meta, data)[0]

    def compress(self, meta, data):
        def get_meta(from_meta, to_meta):
            for key in "ctype", "clevel", "csize":
                if key in from_meta:
                    to_meta[key] = from_meta[key]

        compressor, (cheap_meta, cheap_compressed_data) = self._decide(dict(meta), data)
        if compressor in (LZ4_COMPRESSOR, NONE_COMPRESSOR):
            # we know that trying to compress with expensive compressor is likely pointless,
            # so we fallback to return the cheap compressed data.
            get_meta(cheap_meta, meta)
            return meta, cheap_compressed_data
        # if we get here, the decider decided to try the expensive compressor.
        # we also know that the compressed data returned by the decider is lz4 compressed.
        expensive_meta, expensive_compressed_data = compressor.compress(dict(meta), data)
        ratio = len(expensive_compressed_data) / len(cheap_compressed_data)
        if ratio < 0.99:
            # the expensive compressor managed to squeeze the data significantly better than lz4.
            get_meta(expensive_meta, meta)
            return meta, expensive_compressed_data
        else:
            # otherwise let's just store the lz4 data, which decompresses extremely fast.
            get_meta(cheap_meta, meta)
            return meta, cheap_compressed_data

    def decompress(self, meta, data):
        raise NotImplementedError

    @classmethod
    def detect(cls, data):
        raise NotImplementedError

class ObfuscateSize(CompressorBase):
    """
    Meta-Compressor that obfuscates the compressed data size.
    """
    ID = 0x04
    name = 'obfuscate'

    header_fmt = Struct('<I')
    header_len = len(header_fmt.pack(0))

    def __init__(self, level=None, compressor=None, legacy_mode=False):
        super().__init__(level=level, legacy_mode=legacy_mode)  # data will be encrypted, so we can tell the level
        self.compressor = compressor
        self.level = level
        if level is None:
            pass  # decompression
        elif 1 <= level <= 6:
            self._obfuscate = self._relative_random_reciprocal_obfuscate
            self.factor = 0.001 * 10 ** level
            self.min_r = 0.0001
        elif 110 <= level <= 123:
            self._obfuscate = self._random_padding_obfuscate
            self.max_padding_size = 2 ** (level - 100)  # 1kiB .. 8MiB
        elif level == 250:  # Padmé
            self._obfuscate = self._padme_obfuscate

    def _obfuscate(self, compr_size):
        # implementations need to return the size of obfuscation data,
        # that the caller shall add.
        raise NotImplementedError

    def _relative_random_reciprocal_obfuscate(self, compr_size):
        # effect for SPEC 1:
        # f = 0.01 .. 0.1 for r in 1.0 .. 0.1 == in 90% of cases
        # f = 0.1 .. 1.0 for r in 0.1 .. 0.01 == in 9% of cases
        # f = 1.0 .. 10.0 for r in 0.01 .. 0.001 = in 0.9% of cases
        # f = 10.0 .. 100.0 for r in 0.001 .. 0.0001 == in 0.09% of cases
        r = max(self.min_r, random.random())  # 0..1, but don't get too close to 0
        f = self.factor / r
        return int(compr_size * f)

    def _random_padding_obfuscate(self, compr_size):
        return int(self.max_padding_size * random.random())

    def _padme_obfuscate(self, compr_size):
        if compr_size < 2:
            return 0

        E = math.floor(math.log2(compr_size))  # Get exponent (power of 2)
        S = math.floor(math.log2(E)) + 1       # Second log component
        lastBits = E - S                       # Bits to be zeroed
        bitMask = (2 ** lastBits - 1)          # Mask for rounding

        padded_size = (compr_size + bitMask) & ~bitMask  # Apply rounding

        return padded_size - compr_size  # Return only the additional padding size

    def compress(self, meta, data):
        assert not self.legacy_mode  # we never call this in legacy mode
        meta, compressed_data = self.compressor.compress(meta, data)  # compress data
        compr_size = len(compressed_data)
        assert "csize" in meta, repr(meta)
        meta["psize"] = meta["csize"]  # psize (payload size) is the csize (compressed size) of the inner compressor
        # we only want to obfuscate the size of file content chunks (ROBJ_FILE_STREAM repo objects).
        # for all other types of repo objects (e.g. borg metadata chunks), add 0 length:
        addtl_size = self._obfuscate(compr_size) if meta["type"] == ROBJ_FILE_STREAM else 0
        addtl_size = max(0, addtl_size)  # we can only make it longer, not shorter!
        addtl_size = min(MAX_DATA_SIZE - 1024 - compr_size, addtl_size)  # stay away from MAX_DATA_SIZE
        trailer = bytes(addtl_size)
        obfuscated_data = compressed_data + trailer
        meta["csize"] = len(obfuscated_data)  # csize is the overall output size of this "obfuscation compressor"
        meta["olevel"] = self.level  # remember the obfuscation level, useful for repo-compress
        return meta, obfuscated_data  # for borg2 it is enough that we have the payload size in meta["psize"]

    def decompress(self, meta, data):
        assert self.legacy_mode  # borg2 never dispatches to this, only used for legacy mode
        meta, obfuscated_data = super().decompress(meta, data)  # remove obfuscator ID header
        compr_size = self.header_fmt.unpack(obfuscated_data[0:self.header_len])[0]
        compressed_data = obfuscated_data[self.header_len:self.header_len+compr_size]
        if self.compressor is None:
            compressor_cls = Compressor.detect(compressed_data)[0]
            self.compressor = compressor_cls()
        return self.compressor.decompress(meta, compressed_data)  # decompress data

# Maps valid compressor names to their class
COMPRESSOR_TABLE = {
    CNONE.name: CNONE,
    LZ4.name: LZ4,
    ZLIB.name: ZLIB,
    ZLIB_legacy.name: ZLIB_legacy,
    LZMA.name: LZMA,
    Auto.name: Auto,
    ZSTD.name: ZSTD,
    ObfuscateSize.name: ObfuscateSize,
}
# List of possible compression types. Does not include Auto, since it is a meta-Compressor.
COMPRESSOR_LIST = [LZ4, ZSTD, CNONE, ZLIB, ZLIB_legacy, LZMA, ObfuscateSize, ]  # check fast stuff first

def get_compressor(name, **kwargs):
    cls = COMPRESSOR_TABLE[name]
    return cls(**kwargs)

# compressor instances to be used by all other compressors
NONE_COMPRESSOR = get_compressor('none')
LZ4_COMPRESSOR = get_compressor('lz4')

class Compressor:
    """
    compresses using a compressor with given name and parameters
    decompresses everything we can handle (autodetect)
    """
    def __init__(self, name='null', **kwargs):
        self.params = kwargs
        self.compressor = get_compressor(name, **self.params)

    def compress(self, meta, data):
        return self.compressor.compress(meta, data)

    def decompress(self, meta, data):
        if self.compressor.legacy_mode:
            hdr = data[:2]
        else:
            ctype = meta["ctype"]
            clevel = meta["clevel"]
            hdr = bytes((ctype, clevel))
        compressor_cls = self.detect(hdr)[0]
        return compressor_cls(**self.params).decompress(meta, data)

    @staticmethod
    def detect(data):
        hdr = bytes(data[:2])  # detect() does not work with memoryview
        level = hdr[1]  # usually the level, but not for zlib_legacy
        for cls in COMPRESSOR_LIST:
            if cls.detect(hdr):
                return cls, (255 if cls.name == 'zlib_legacy' else level)
        else:
            raise ValueError('No decompressor for this data found: %r.', data[:2])



import logging

from ._common import with_repository
from ..constants import *  # NOQA
from ..helpers import format_archive, CommandError, bin_to_hex, archivename_validator
from ..helpers.argparsing import ArgumentParser
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


class UnDeleteMixIn:
    @with_repository(manifest=False)
    def do_undelete(self, args, repository):
        """Undeletes archives."""
        self.output_list = args.output_list
        dry_run = args.dry_run
        manifest = Manifest.load(repository, (Manifest.Operation.DELETE,))
        if args.name:
            archive_infos = [manifest.archives.get_one([args.name], deleted=True)]
        else:
            args.deleted = True
            archive_infos = manifest.archives.list_considering(args)
        count = len(archive_infos)
        if count == 0:
            return
        if not args.name and not args.match_archives and args.first == 0 and args.last == 0:
            raise CommandError("Aborting: if you really want to undelete all archives, please use -a 'sh:*'.")

        undeleted = False
        logger_list = logging.getLogger("borg.output.list")
        for i, archive_info in enumerate(archive_infos, 1):
            name, id, hex_id = archive_info.name, archive_info.id, bin_to_hex(archive_info.id)
            try:
                if not dry_run:
                    manifest.archives.undelete_by_id(id)
            except KeyError:
                self.print_warning(f"Archive {name} {hex_id} not found ({i}/{count}).")
            else:
                undeleted = True
                if self.output_list:
                    msg = "Would undelete: {} ({}/{})" if dry_run else "Undeleted archive: {} ({}/{})"
                    logger_list.info(msg.format(format_archive(archive_info), i, count))
        if dry_run:
            logger.info("Finished dry-run.")
        elif undeleted:
            manifest.write()
            self.print_warning("Done.", wc=None)
        else:
            self.print_warning("Aborted.", wc=None)
        return

    def build_parser_undelete(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog, define_archive_filters_group

        undelete_epilog = process_epilog(
            """
        This command undeletes archives in the repository.

        Important: Undeleting archives is only possible before compacting.
        Once ``borg compact`` has run, all disk space occupied only by the
        soft-deleted archives will be freed, and undeleting is no longer
        possible.

        When in doubt, use ``--dry-run --list`` to see what would be
        undeleted.

        You can undelete multiple archives by specifying a match pattern using
        the ``--match-archives PATTERN`` option (for more information on these
        patterns, see :ref:`borg_patterns`).
        """
        )
        subparser = ArgumentParser(
            parents=[common_parser], description=self.do_undelete.__doc__, epilog=undelete_epilog
        )
        subparsers.add_subcommand("undelete", subparser, help="undelete archives")
        subparser.add_argument(
            "-n", "--dry-run", dest="dry_run", action="store_true", help="do not change the repository"
        )
        subparser.add_argument(
            "--list", dest="output_list", action="store_true", help="output a verbose list of archives"
        )
        define_archive_filters_group(subparser)
        subparser.add_argument(
            "name", metavar="NAME", nargs="?", type=archivename_validator, help="specify the archive name"
        )



from hashlib import sha256
from io import BytesIO
import os
import random

import pytest

from . import cf, cf_expand
from ...chunkers import ChunkerFastCDC, get_chunker
from ...chunkers.fastcdc import fastcdc_get_gear_table
from ...constants import *  # NOQA
from ...helpers import hex_to_bin


# from os.urandom(32)
key0 = hex_to_bin("ad9f89095817f0566337dc9ee292fcd59b70f054a8200151f1df5f21704824da")
key1 = hex_to_bin("f1088c7e9e6ae83557ad1558ff36c44a369ea719d1081c29684f52ffccb72cb8")


def H(data):
    return sha256(data).digest()


def test_chunkpoints_fastcdc_unchanged():
    def twist(size):
        x = 1
        a = bytearray(size)
        for i in range(size):
            x = (x * 1103515245 + 12345) & 0x7FFFFFFF
            a[i] = x & 0xFF
        return a

    data = twist(100000)

    runs = []
    for nc_level in (0, 2, 3):
        for minexp in (4, 6, 7, 11, 12):
            for maxexp in (15, 17):
                if minexp >= maxexp:
                    continue
                for maskbits in (4, 7, 10, 12):
                    if maskbits - nc_level < 1:  # nc_level needs room below the base mask bits
                        continue
                    for key in (key0, key1):
                        fh = BytesIO(data)
                        chunker = ChunkerFastCDC(key, minexp, maxexp, maskbits, nc_level)
                        chunks = [H(c) for c in cf(chunker.chunkify(fh, -1))]
                        runs.append(H(b"".join(chunks)))

    # The "correct" hash below matches the existing chunker behavior.
    # Future chunker optimizations must not change this, or existing repos will bloat.
    overall_hash = H(b"".join(runs))
    print(overall_hash.hex())
    assert overall_hash == hex_to_bin("50d39b6f30214d78f665ff97a4800142cddcb6a7c5995e5d162f9c6dceb20cfe")


def test_fastcdc_chunksize_distribution():
    data = os.urandom(1048576)
    min_exp, max_exp, mask, nc_level = 10, 16, 14, 2  # chunk size target 16 KiB, clip at 1 KiB and 64 KiB
    chunker = ChunkerFastCDC(key0, min_exp, max_exp, mask, nc_level)
    f = BytesIO(data)
    chunks = cf(chunker.chunkify(f))
    del chunks[-1]  # get rid of the last chunk, it can be smaller than 2**min_exp
    chunk_sizes = [len(chunk) for chunk in chunks]
    chunks_count = len(chunks)
    min_chunksize_observed = min(chunk_sizes)
    max_chunksize_observed = max(chunk_sizes)
    min_count = sum(int(size == 2**min_exp) for size in chunk_sizes)
    max_count = sum(int(size == 2**max_exp) for size in chunk_sizes)
    print(
        f"count: {chunks_count} min: {min_chunksize_observed} max: {max_chunksize_observed} "
        f"min count: {min_count} max count: {max_count}"
    )
    # usually there will about 64 chunks
    assert 32 < chunks_count < 128
    # chunks always must be between min and max (clipping must work):
    assert min_chunksize_observed >= 2**min_exp
    assert max_chunksize_observed <= 2**max_exp
    # most chunks should be cut due to the gear hash triggering, not due to clipping at min/max size:
    assert min_count < 10
    assert max_count < 10


def test_fastcdc_gear_table():
    # Test that the function returns a list of 256 integers
    table0 = fastcdc_get_gear_table(key0)
    assert len(table0) == 256
    for value in table0:
        assert isinstance(value, int)
        assert 0 <= value < 2**64

    # deterministic (same key produces same table)
    assert table0 == fastcdc_get_gear_table(key0)

    # different keys produce different tables
    table1 = fastcdc_get_gear_table(key1)
    assert table0 != table1


def test_fastcdc_get_chunker():
    # without a key, get_chunker uses an all-zero key; chunking must still work and be deterministic
    data = os.urandom(2 * 1024 * 1024)
    a = cf_expand(get_chunker(*FASTCDC_PARAMS, key=None).chunkify(BytesIO(data)))
    b = cf_expand(get_chunker(*FASTCDC_PARAMS, key=None).chunkify(BytesIO(data)))
    assert a == b
    assert b"".join(a) == data


def test_fastcdc_params_parsing():
    from argparse import ArgumentTypeError

    from ...helpers import ChunkerParams

    # fastcdc, chunk_min, chunk_max, chunk_mask, nc_level (no window field)
    assert ChunkerParams("fastcdc,19,23,21,2") == (CH_FASTCDC, 19, 23, 21, 2)
    assert ChunkerParams("fastcdc,10,23,16,0") == (CH_FASTCDC, 10, 23, 16, 0)
    # a 6-field (buzhash64-style, with window) fastcdc must be rejected
    with pytest.raises(ArgumentTypeError):
        ChunkerParams("fastcdc,19,23,21,4095,2")
    # a 4-field fastcdc (missing nc_level) must be rejected, not fall into old-style compat mode
    with pytest.raises(ArgumentTypeError):
        ChunkerParams("fastcdc,19,23,21")
    # nc_level out of range (chunk_mask - nc_level < 1)
    with pytest.raises(ArgumentTypeError):
        ChunkerParams("fastcdc,19,23,21,21")
    # chunk_min <= chunk_mask <= chunk_max violated
    with pytest.raises(ArgumentTypeError):
        ChunkerParams("fastcdc,19,23,24,2")
    # chunk_min == chunk_max: the chunker needs min_size + 1 <= max_size
    with pytest.raises(ArgumentTypeError):
        ChunkerParams("fastcdc,20,20,20,2")


@pytest.mark.skipif("BORG_TESTS_SLOW" not in os.environ, reason="slow tests not enabled, use BORG_TESTS_SLOW=1")
@pytest.mark.parametrize("worker", range(os.cpu_count() or 1))
def test_fuzz_fastcdc(worker):
    # Fuzz fastcdc with random and uniform data of misc. sizes and misc keys.
    def rnd_key():
        return os.urandom(32)

    # decompose FASTCDC_PARAMS = (algo, min_exp, max_exp, mask_bits, nc_level)
    algo, min_exp, max_exp, mask_bits, nc_level = FASTCDC_PARAMS
    assert algo == CH_FASTCDC

    keys = [b"\0" * 32] + [rnd_key() for _ in range(10)]
    sizes = [random.randint(1, 4 * 1024 * 1024) for _ in range(50)]

    for key in keys:
        chunker = ChunkerFastCDC(key, min_exp, max_exp, mask_bits, nc_level)
        for size in sizes:
            # Random data
            data = os.urandom(size)
            with BytesIO(data) as bio:
                parts = cf_expand(chunker.chunkify(bio))
            assert b"".join(parts) == data

            # All-same data (non-zero)
            data = b"\x42" * size
            with BytesIO(data) as bio:
                parts = cf_expand(chunker.chunkify(bio))
            assert b"".join(parts) == data

            # All-zero data
            data = b"\x00" * size
            with BytesIO(data) as bio:
                parts = cf_expand(chunker.chunkify(bio))
            assert b"".join(parts) == data



from typing import Callable, NamedTuple
from datetime import datetime, timedelta
import logging
import math
from functools import wraps
import os
from itertools import count, combinations
from ._common import with_repository, Highlander
from ..constants import *  # NOQA
from ..helpers import ArchiveFormatter, ProgressIndicatorPercent, CommandError, Error
from ..helpers import archivename_validator, int_or_interval, sig_int, timestamp
from ..helpers import json_print, basic_json_data
from ..helpers.argparsing import ArgumentParser
from ..manifest import ArchiveInfo, Manifest

from ..logger import create_logger

logger = create_logger()


class PruningRule(NamedTuple):
    key: str
    period_func: Callable[[ArchiveInfo | datetime], str]

    def __str__(self):
        return self.key


class KeepResult(NamedTuple):
    rule: PruningRule
    idx: int
    oldest: bool = False

    def __str__(self):
        return f"Keep(rule={self.rule}, idx={self.idx}{', oldest=True' if self.oldest else ''})"


def archive_datetime_dispatch(func: Callable[[datetime], str]) -> Callable[[ArchiveInfo | datetime], str]:
    """
    Wraps a datetime-taking function with a dispatcher that can call that
    function by extracting the timestamp from an archive.
    """

    @wraps(func)
    def wrapper(arg):
        if isinstance(arg, datetime):
            return func(arg)
        if isinstance(arg, ArchiveInfo):
            return func(arg.ts)
        raise TypeError(f"{func.__name__}(): expected datetime or Archive, " f"got {type(arg).__name__}")

    return wrapper


# The *_period_func group of functions create period grouping keys to group
# together archives falling within a certain period. Among archives in each of
# these groups, only the latest (by creation timestamp) is kept.


def unique_period_func():
    counter = count()
    max_digits = math.ceil(math.log10(MAX_ARCHIVES))

    @archive_datetime_dispatch
    def unique_values(_dt):
        """Group archives by an incrementing counter, practically making each archive a group of 1"""
        # zfill ensures lexicographic ordering matches number ordering in case of comparisons
        return str(next(counter)).zfill(max_digits)

    return unique_values


def pattern_period_func(pattern):
    @archive_datetime_dispatch
    def inner(dt):
        """Group archives by extracting given strftime-pattern from their creation timestamp"""
        # compute in local timezone
        return dt.astimezone().strftime(pattern)

    return inner


@archive_datetime_dispatch
def quarterly_13weekly_period_func(dt):
    """Group archives by extracting the ISO-8601 13-week quarter from their creation timestamp"""
    (year, week) = dt.astimezone().isocalendar()[:2]  # local time
    return f"{year}-{min(max((week - 1) // 13, 0), 3):02}"


@archive_datetime_dispatch
def quarterly_3monthly_period_func(dt):
    """Group archives by extracting the 3-month quarter from their creation timestamp"""
    (year, month) = dt.astimezone().timetuple()[:2]  # local time
    return f"{year}-{(month - 1) // 3:02}"


# Each archive is considered for keeping
PRUNE_KEEP = PruningRule("keep", unique_period_func())
# Last archive (by creation timestamp) within period group is considered for keeping
PRUNE_SECONDLY = PruningRule("secondly", pattern_period_func("%Y-%m-%d %H:%M:%S"))
PRUNE_MINUTELY = PruningRule("minutely", pattern_period_func("%Y-%m-%d %H:%M"))
PRUNE_HOURLY = PruningRule("hourly", pattern_period_func("%Y-%m-%d %H"))
PRUNE_DAILY = PruningRule("daily", pattern_period_func("%Y-%m-%d"))
PRUNE_WEEKLY = PruningRule("weekly", pattern_period_func("%G-%V"))
PRUNE_MONTHLY = PruningRule("monthly", pattern_period_func("%Y-%m"))
PRUNE_QUARTERLY_13WEEKLY = PruningRule("quarterly_13weekly", quarterly_13weekly_period_func)
PRUNE_QUARTERLY_3MONTHLY = PruningRule("quarterly_3monthly", quarterly_3monthly_period_func)
PRUNE_YEARLY = PruningRule("yearly", pattern_period_func("%Y"))

# Fake rule used to indicate archives skipped by --from
PRUNE_FROM = PruningRule("skip", unique_period_func())

PRUNING_RULES = [
    PRUNE_KEEP,
    PRUNE_SECONDLY,
    PRUNE_MINUTELY,
    PRUNE_HOURLY,
    PRUNE_DAILY,
    PRUNE_WEEKLY,
    PRUNE_MONTHLY,
    PRUNE_QUARTERLY_13WEEKLY,
    PRUNE_QUARTERLY_3MONTHLY,
    PRUNE_YEARLY,
]


def prune(
    archives: list[ArchiveInfo],
    rule: PruningRule,
    n_or_interval: int | timedelta,
    base_timestamp: datetime | None,
    keep_oldest: bool,
    previously_kept: frozenset[ArchiveInfo] = frozenset(),
) -> dict[ArchiveInfo, KeepResult]:
    if len(archives) == 0 or n_or_interval in (0, timedelta(0)):
        return {}

    if isinstance(n_or_interval, int):
        n, earliest_timestamp = n_or_interval, None
    else:
        if base_timestamp is None:
            raise ValueError("base_timestamp is required when using interval-based pruning")
        n, earliest_timestamp = None, base_timestamp - n_or_interval

    keep: dict[ArchiveInfo, KeepResult] = {}

    def can_retain(a):
        if n is not None:
            return n == -1 or len(keep) < n
        else:
            return a.ts >= earliest_timestamp

    prev_period = None
    for archive in archives:
        if not can_retain(archive):
            break
        period = rule.period_func(archive)
        if period != prev_period:
            prev_period = period
            if archive not in keep and archive not in previously_kept:
                keep[archive] = KeepResult(rule=rule, idx=len(keep))

    if keep_oldest:
        # Keep oldest archive if we didn't reach the target retention.
        oldest_archive = archives[-1]
        if oldest_archive not in keep and oldest_archive not in previously_kept and can_retain(oldest_archive):
            keep[oldest_archive] = KeepResult(rule=rule, idx=len(keep), oldest=True)

    return keep


class PruneMixIn:
    @with_repository(compatibility=(Manifest.Operation.DELETE,))
    def do_prune(self, args, repository, manifest):
        """Prune archives according to specified rules."""
        self._validate_prune_args(args)

        match = [args.name] if args.name else args.match_archives
        archives = manifest.archives.list(match=match, sort_by=["ts"], reverse=True)
        archives = [ai for ai in archives if "@PROT" not in ai.tags]

        # Archives to keep along with the rule that ensured them being kept
        keep = {}

        from_timestamp = getattr(args, PRUNE_FROM.key)
        candidate_archives = archives

        if from_timestamp is not None:
            base_timestamp = from_timestamp

            # `--from` is a prefilter: Archives made at or after this time are kept by default. They are not considered
            # for pruning at all and thus won't falsely occupy an active retention period.
            for archive in archives:
                if archive.ts < from_timestamp:
                    break
                keep[archive] = KeepResult(rule=PRUNE_FROM, idx=len(keep))
            candidate_archives = archives[len(keep) :]
        else:
            base_timestamp = datetime.now().astimezone()

        # Apply each retention rule to all candidate archives. The
        # `previously_kept` parameter prevents later (coarser-grained) rules
        # from double-counting archives already retained by earlier rules.
        active_rules = [
            (rule, getattr(args, rule.key)) for rule in PRUNING_RULES if getattr(args, rule.key) is not None
        ]
        for rule, n_or_interval in active_rules:
            keep |= prune(
                archives=candidate_archives,
                rule=rule,
                n_or_interval=n_or_interval,
                base_timestamp=base_timestamp,
                keep_oldest=(
                    rule == active_rules[-1][0]
                ),  # Activate keep_oldest rule only for the largest active interval
                previously_kept=frozenset(keep),
            )

        archives_to_prune = set(archives) - set(keep)

        if args.format is not None:
            format = args.format
        elif args.short:
            format = "{archive}"
        else:
            format = os.environ.get("BORG_PRUNE_FORMAT", "{archive:<36} {time} [{id}]")
        formatter = ArchiveFormatter(format, repository, manifest, manifest.key, iec=args.iec)

        if args.json:
            output_data = []
        else:
            logger.info("Repository contains %d archives.", manifest.archives.count())
            logger.info("Applying rules to the matching %d archives...", len(archives))
            logger.info("Keeping %d archives, pruning %d archives.", len(keep), len(archives_to_prune))

        list_logger = logging.getLogger("borg.output.list")
        # set up counters for the progress display
        num_archives_deleted = 0
        pi = ProgressIndicatorPercent(total=len(archives_to_prune), msg="Pruning archives %3.0f%%", msgid="prune")
        for archive_info in archives:
            if sig_int and sig_int.action_done():
                break
            # get_item_data/format_item may internally load the archive from the repository,
            # so we must call it before deleting the archive.
            if args.json:
                archive_data = formatter.get_item_data(archive_info, jsonline=True)
            else:
                archive_formatted = formatter.format_item(archive_info, jsonline=False)
            if archive_info in archives_to_prune:
                if not args.json:
                    pi.show()
                num_archives_deleted += 1
                if args.dry_run:
                    log_message = "Would prune:"
                else:
                    log_message = f"Pruning archive ({num_archives_deleted}/{len(archives_to_prune)}):"
                    manifest.archives.delete_by_id(archive_info.id)
                if args.json:
                    archive_data["kept"] = False
                    archive_data["deleted_archive_number"] = num_archives_deleted
            else:
                result = keep[archive_info]
                result_message = f"{result.rule.key}{'[oldest]' if result.oldest else ''} #{result.idx + 1}"
                log_message = f"Keeping archive (rule: {result_message}):"
                if args.json:
                    archive_data["kept"] = True
                    archive_data["keep_rule"] = result.rule.key
                    archive_data["kept_oldest"] = result.oldest
                    archive_data["kept_archive_number"] = result.idx + 1
            if args.json:
                if (
                    args.output_list
                    or not (args.list_pruned or args.list_kept)
                    or (args.list_pruned and archive_info in archives_to_prune)
                    or (args.list_kept and archive_info not in archives_to_prune)
                ):
                    output_data.append(archive_data)
            elif (
                args.output_list
                or (args.list_pruned and archive_info in archives_to_prune)
                or (args.list_kept and archive_info not in archives_to_prune)
            ):
                list_logger.info(f"{log_message:<44} {archive_formatted}")
        if not args.json:
            pi.finish()
        if args.json:
            json_print(basic_json_data(manifest, extra={"archives": output_data}))
        if num_archives_deleted > 0 and not args.dry_run:
            manifest.write()
            self.print_warning('Done. Run "borg compact" to free space.', wc=None)
        if sig_int:
            raise Error("Got Ctrl-C / SIGINT.")

    def _validate_prune_args(self, args):
        keep_args = {rule.key: getattr(args, rule.key) for rule in PRUNING_RULES if getattr(args, rule.key) is not None}

        if len(keep_args) == 0:
            raise CommandError(
                'At least one of the "keep", "keep-secondly", "keep-minutely", "keep-hourly", "keep-daily", '
                '"keep-weekly", "keep-monthly", "keep-13weekly", "keep-3monthly", or "keep-yearly" settings must be '
                "specified."
            )

        if all(not bool(val) for val in keep_args.values()):
            raise CommandError(
                'None of the "keep", "keep-secondly", "keep-minutely", "keep-hourly", "keep-daily", "keep-weekly", '
                '"keep-monthly", "keep-13weekly", "keep-3monthly", or "keep-yearly" settings have a positive value. '
                "At least one must be non-zero."
            )

        def lo_hi_mismatch_errmsg(lo_arg, lo_val, hi_arg, hi_val):
            return (
                f"The combination of \"{lo_arg}='{lo_val}'\" and \"{hi_arg}='{hi_val}'\" is invalid. It is effectively "
                f"useless since every archive matched by {hi_arg} would have already been matched by {lo_arg}."
            )

        interval_args = [(arg, val) for arg, val in keep_args.items() if isinstance(val, timedelta) or val == -1]
        for (lo_arg, lo_val), (hi_arg, hi_val) in combinations(interval_args, 2):
            if hi_val == -1:
                # 'Infinity' is always bigger
                continue

            if lo_val == -1 or lo_val >= hi_val:
                raise CommandError(lo_hi_mismatch_errmsg(lo_arg, lo_val, hi_arg, hi_val))

        int_args = [(arg, val) for arg, val in keep_args.items() if isinstance(val, int)]
        for (lo_arg, lo_val), (hi_arg, hi_val) in combinations(int_args, 2):
            if lo_val == -1:
                raise CommandError(lo_hi_mismatch_errmsg(lo_arg, lo_val, hi_arg, hi_val))

    def build_parser_prune(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog
        from ._common import define_archive_filters_group

        prune_epilog = process_epilog(
            """
        The prune command prunes a repository by soft-deleting all archives not
        matching any of the specified retention options.

        Important:

        - The prune command will only mark archives for deletion ("soft-deletion"),
          repository disk space is **not** freed until you run ``borg compact``.
        - You can use ``borg undelete`` to undelete archives, but only until
          you run ``borg compact``.

        This command is normally used by automated backup scripts wanting to keep a
        certain number of historic backups. This retention policy is commonly referred to as
        `GFS <https://en.wikipedia.org/wiki/Backup_rotation_scheme#Grandfather-father-son>`_
        (Grandfather-father-son) backup rotation scheme.

        The recommended way to use prune is to give the archive series name to it via the
        NAME argument (assuming you have the same name for all archives in a series).
        Alternatively, you can also use --match-archives (-a), then only archives that
        match the pattern are considered for deletion and only those archives count
        towards the totals specified by the rules.
        Otherwise, *all* archives in the repository are candidates for deletion!
        There is no automatic distinction between archives representing different
        contents. These need to be distinguished by specifying matching globs.

        If you have multiple series of archives with different data sets (e.g.
        from different machines) in one shared repository, use one prune call per
        series.

        The ``--keep`` option is the simplest way to specify a basic retention
        policy. It accepts a count or a time interval for retention (e.g.
        ``10`` or ``7d``, ``4w``). With a count it keeps at most that many
        recent archives; with an interval it keeps all archives created within
        that time window. When ``--from`` is given together with an interval
        retention, the interval is measured backwards from that timestamp
        instead of from the current time. See ``Date and Time`` docs for exact
        INTERVAL format.

        The ``--keep-secondly``, ``--keep-minutely``, ``--keep-hourly``,
        ``--keep-daily``, ``--keep-weekly``, ``--keep-monthly``,
        ``--keep-13weekly``, ``--keep-3monthly``, and ``--keep-yearly`` options
        specify time period retention policies. They accept either a count N for
        retention or a time interval INTERVAL for retention, same as for ``--keep``.
        With a retention count, they keep at most that many archives (one per
        period, e.g. one per day or one per month until the retention count is
        met). With a retention interval, they keep one archive per period
        within that time span (e.g. at most one per day in a span of seven
        days, even if some days had none) -- measured from ``--from`` if given,
        otherwise from the current time. Specifying a count of ``-1`` (or the
        word ``all``) means no limit. A zero count or zero-length interval
        keeps nothing.

        The ``--from`` option restricts pruning to archives older than the given
        TIMESTAMP. Archives made at or after this timestamp are kept unconditionally
        as a pre-filter. When ``--from`` is used together with interval-based
        ``--keep-*`` options (e.g. ``--keep-daily 7d``), the interval is measured
        backwards from the given timestamp rather than from the current time.
        Count-based retention does not count the unconditionally kept archives.

        The ``--keep-13weekly`` and ``--keep-3monthly`` rules are two different
        strategies for keeping archives every quarter year.

        The oldest archive is kept as long as the coarsest retention rule covers it --
        ``--keep-yearly=3`` will keep the oldest archive if it couldn't otherwise find
        three candidates, ``--keep-yearly=5y`` will keep the oldest archive as long as
        it is at or within the 5y interval. This is useful for rolling tiered backup
        schemes, where the earliest backup in a retention window should survive until
        the next tier's interval naturally replaces it.

        When using interval-based pruning with multiple ``--keep-*`` options,
        the intervals must be specified in increasing length matching the
        periods chosen. For example, ``--keep-daily 7d --keep-weekly 4w`` is
        valid, but ``--keep-daily 30d --keep-weekly 7d`` is not, because the
        weekly interval is already covered by the daily one and so the weekly
        interval is effectively useless. An error is emitted upon running
        ``borg prune`` if such a combination of flags is given. The order of
        flags on the command line is not significant.


        A practical approach for recurring backups is to use rules
        with increasing coarseness so that most of recent history is kept and
        older history gradually thins out with time. For example,
        ``--keep-daily 7d --keep-weekly 4w --keep-monthly 6`` keeps an
        archive per day for the past week, per week for the past month, and
        one per month for six months after that. Combine this with ``--from``
        to align time windows to calendar boundaries rather than the exact
        moment you run prune for more predictable behavior of coarser rules:
        ``--keep-daily 7d --keep-weekly 4w --from $(date +%F)``.

        Count-based retention keeps archives less bound to time. For instance,
        ``--keep-yearly 3`` retains 3 yearly archives however far back they
        span and ``--keep-daily 20`` keeps 20 archives no matter if you missed
        a week in between. This can be useful for less regular archive
        creation, or if your use case does not map well to specific time
        intervals, or if you simply prefer to think of archive retention in
        numbers rather than intervals.

        For count-based retention, backups selected by more granular rules do
        not count towards those of coarser rules. ``--keep 3 --keep-monthly 2``
        will first keep the 3 latest archives and then keep 2 monthly archives,
        skipping ones that were already kept by ``--keep 3``.

        The time that each archive creation started is used to match archives
        to pruning periods. Dates and times are interpreted in the local
        timezone of your system. Weeks go from Monday to Sunday.


        You can influence how the ``--list`` output is formatted by using the ``--short``
        option (less wide output) or by giving a custom format using ``--format`` (see
        the ``borg repo-list`` description for more details about the format string).
        """
        )
        subparser = ArgumentParser(parents=[common_parser], description=self.do_prune.__doc__, epilog=prune_epilog)
        subparsers.add_subcommand("prune", subparser, help="prune archives")
        subparser.add_argument(
            "-n", "--dry-run", dest="dry_run", action="store_true", help="do not change the repository"
        )
        subparser.add_argument(
            "--list", dest="output_list", action="store_true", help="output a verbose list of archives it keeps/prunes"
        )
        subparser.add_argument("--short", dest="short", action="store_true", help="use a less wide archive part format")
        subparser.add_argument(
            "--list-pruned", dest="list_pruned", action="store_true", help="output verbose list of archives it prunes"
        )
        subparser.add_argument(
            "--list-kept", dest="list_kept", action="store_true", help="output verbose list of archives it keeps"
        )
        subparser.add_argument(
            "--format",
            metavar="FORMAT",
            dest="format",
            action=Highlander,
            help="specify format for the archive part " '(default: "{archive:<36} {time} [{id}]")',
        )
        subparser.add_argument(
            "--json",
            action="store_true",
            help="Format output as JSON. "
            "The form of ``--format`` is ignored, "
            "but keys used in it are added to the JSON output. "
            "Some keys are always present. Note: JSON can only represent text.",
        )
        subparser.add_argument(
            "--from",
            metavar="TIMESTAMP",
            dest=PRUNE_FROM.key,
            type=timestamp,
            action=Highlander,
            help="only consider archives older than this for pruning",
        )
        subparser.add_argument(
            "--keep",
            dest=PRUNE_KEEP.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of archives to keep",
        )
        subparser.add_argument(
            "--keep-secondly",
            dest=PRUNE_SECONDLY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of secondly archives to keep",
        )
        subparser.add_argument(
            "--keep-minutely",
            dest=PRUNE_MINUTELY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of minutely archives to keep",
        )
        subparser.add_argument(
            "-H",
            "--keep-hourly",
            dest=PRUNE_HOURLY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of hourly archives to keep",
        )
        subparser.add_argument(
            "-d",
            "--keep-daily",
            dest=PRUNE_DAILY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of daily archives to keep",
        )
        subparser.add_argument(
            "-w",
            "--keep-weekly",
            dest=PRUNE_WEEKLY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of weekly archives to keep",
        )
        subparser.add_argument(
            "-m",
            "--keep-monthly",
            dest=PRUNE_MONTHLY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of monthly archives to keep",
        )
        quarterly_group = subparser.add_mutually_exclusive_group()
        quarterly_group.add_argument(
            "--keep-13weekly",
            dest=PRUNE_QUARTERLY_13WEEKLY.key,
            type=int_or_interval,
            help="number or time interval of quarterly archives to keep (13 week strategy)",
        )
        quarterly_group.add_argument(
            "--keep-3monthly",
            dest=PRUNE_QUARTERLY_3MONTHLY.key,
            type=int_or_interval,
            help="number or time interval of quarterly archives to keep (3 month strategy)",
        )
        subparser.add_argument(
            "-y",
            "--keep-yearly",
            dest=PRUNE_YEARLY.key,
            type=int_or_interval,
            action=Highlander,
            help="number or time interval of yearly archives to keep",
        )
        define_archive_filters_group(subparser, sort_by=False, first_last=False)
        subparser.add_argument(
            "name", metavar="NAME", nargs="?", type=archivename_validator, help="specify the archive name"
        )



import hmac
import os
from hashlib import pbkdf2_hmac

from ...constants import *  # NOQA
from ...crypto.low_level import AES256_CTR_HMAC_SHA256, AES256_CTR_BLAKE2b, hmac_sha256, blake2b_256
from ...crypto.key import ID_HMAC_SHA_256, AESKeyBase, FlexiKey, AuthenticatedKeyBase, UnsupportedKeyFormatError
from ...helpers import get_limited_unpacker, msgpack
from ...item import EncryptedKey
from .low_level import AES


class Pbkdf2FileMixin:
    """Mixin for borg 1.x key files encrypted with PBKDF2 + AES-CTR."""

    @staticmethod
    def pbkdf2(passphrase, salt, iterations, output_len_in_bytes):
        if os.environ.get("BORG_TESTONLY_WEAKEN_KDF") == "1":
            iterations = 1
        return pbkdf2_hmac("sha256", passphrase.encode("utf-8"), salt, iterations, output_len_in_bytes)

    def decrypt_key_file(self, data, passphrase):
        unpacker = get_limited_unpacker("key")
        unpacker.feed(data)
        unpacked = unpacker.unpack()
        encrypted_key = EncryptedKey(internal_dict=unpacked)
        if encrypted_key.version != 1:
            raise UnsupportedKeyFormatError()
        self._encrypted_key_algorithm = encrypted_key.algorithm
        if encrypted_key.algorithm == "sha256":
            return self.decrypt_key_file_pbkdf2(encrypted_key, passphrase)
        return super().decrypt_key_file(data, passphrase)

    def encrypt_key_file(self, data, passphrase, algorithm, label=None):
        if algorithm == "sha256":
            return self.encrypt_key_file_pbkdf2(data, passphrase)
        return super().encrypt_key_file(data, passphrase, algorithm, label=label)

    def decrypt_key_file_pbkdf2(self, encrypted_key, passphrase):
        key = self.pbkdf2(passphrase, encrypted_key.salt, encrypted_key.iterations, 32)
        data = AES(key, b"\0" * 16).decrypt(encrypted_key.data)
        if hmac.compare_digest(hmac_sha256(key, data), encrypted_key.hash):
            return data
        return None

    def encrypt_key_file_pbkdf2(self, data, passphrase):
        salt = os.urandom(32)
        iterations = PBKDF2_ITERATIONS
        key = self.pbkdf2(passphrase, salt, iterations, 32)
        hash = hmac_sha256(key, data)
        cdata = AES(key, b"\0" * 16).encrypt(data)
        enc_key = EncryptedKey(version=1, salt=salt, iterations=iterations, algorithm="sha256", hash=hash, data=cdata)
        return msgpack.packb(enc_key.as_dict())


def random_blake2b_256_key():
    # This might look a bit curious, but is the same construction used in the keyed mode of BLAKE2b.
    # Why limit the key to 64 bytes and pad it with 64 nulls nonetheless? The answer is that BLAKE2b
    # has a 128 byte block size, but only 64 bytes of internal state (this is also referred to as a
    # "local wide pipe" design, because the compression function transforms (block, state) => state,
    # and len(block) >= len(state), hence wide.)
    # In other words, a key longer than 64 bytes would have simply no advantage, since the function
    # has no way of propagating more than 64 bytes of entropy internally.
    # It's padded to a full block so that the key is never buffered internally by blake2b_update, ie.
    # it remains in a single memory location that can be tracked and could be erased securely, if we
    # wanted to.
    return os.urandom(64) + bytes(64)


class ID_BLAKE2b_256:
    """
    Key mix-in class for using BLAKE2b-256 for the id key.

    The id_key length must be 32 bytes.
    """

    IDHASH_NAME = "blake2"

    def id_hash(self, data):
        return blake2b_256(self.id_key, data)

    def init_from_random_data(self):
        super().init_from_random_data()
        enc_key = os.urandom(32)
        enc_hmac_key = random_blake2b_256_key()
        self.crypt_key = enc_key + enc_hmac_key
        self.id_key = random_blake2b_256_key()


class Blake2AuthenticatedKey(ID_BLAKE2b_256, AuthenticatedKeyBase):  # type: ignore[misc]
    TYPE = KeyType.BLAKE2AUTHENTICATED
    TYPES_ACCEPTABLE = {TYPE}
    ENC_NAME = "authenticated"  # IDHASH_NAME = "blake2" via ID_BLAKE2b_256 mix-in; read-only (borg 1.x)


# borg 1.x AES-CTR keys. keyfile and repokey are no longer separate classes - storage is a per-key
# property (self.storage), tracked when the key is loaded. These classes are read-only (borg 2 only
# reads borg 1.x repos, e.g. via borg transfer; it never creates them), so the canonical TYPE byte is
# never written - only TYPES_ACCEPTABLE matters, and it keeps the historic keyfile/repokey/passphrase bytes.


class AESCTRKey(Pbkdf2FileMixin, ID_HMAC_SHA_256, AESKeyBase, FlexiKey):  # type: ignore[misc]
    TYPES_ACCEPTABLE = {KeyType.KEYFILE, KeyType.REPO, KeyType.PASSPHRASE}
    TYPE = KeyType.KEYFILE
    ENC_NAME = "aes256-ctr"  # IDHASH_NAME = "sha256" via ID_HMAC_SHA_256 mix-in; read-only (borg 1.x)
    STORAGE = KeyBlobStorage.REPO  # seed default; actual per-key storage is tracked in self.storage on load
    LOCATION_CONFIGURABLE = True  # borg 1.x had keyfile and repokey variants
    CIPHERSUITE = AES256_CTR_HMAC_SHA256


class Blake2AESCTRKey(Pbkdf2FileMixin, ID_BLAKE2b_256, AESKeyBase, FlexiKey):  # type: ignore[misc]
    TYPES_ACCEPTABLE = {KeyType.BLAKE2KEYFILE, KeyType.BLAKE2REPO}
    TYPE = KeyType.BLAKE2KEYFILE
    ENC_NAME = "aes256-ctr"  # IDHASH_NAME = "blake2" via ID_BLAKE2b_256 mix-in; read-only (borg 1.x)
    STORAGE = KeyBlobStorage.REPO  # seed default; actual per-key storage is tracked in self.storage on load
    LOCATION_CONFIGURABLE = True  # borg 1.x had keyfile and repokey variants
    CIPHERSUITE = AES256_CTR_BLAKE2b


LEGACY_KEY_TYPES = (AESCTRKey, Blake2AESCTRKey, Blake2AuthenticatedKey)



import functools
import os
import subprocess
import tempfile

import pytest

from . import cmd, generate_archiver_tests, RK_ENCRYPTION

pytest_generate_tests = lambda metafunc: generate_archiver_tests(metafunc, kinds="local")  # NOQA


@functools.lru_cache
def cmd_available(cmd):
    """Check if a shell command is available."""
    try:
        subprocess.run(cmd.split(), capture_output=True, check=True)
        return True
    except (subprocess.SubprocessError, FileNotFoundError):
        return False


needs_bash = pytest.mark.skipif(not cmd_available("bash --version"), reason="Bash not available")
needs_zsh = pytest.mark.skipif(not cmd_available("zsh --version"), reason="Zsh not available")


def _run_bash_completion_fn(completion_script, setup_code):
    """Source the completion script in bash and run setup_code, return subprocess result."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".bash", delete=False) as f:
        f.write(completion_script)
        script_path = f.name
    try:
        result = subprocess.run(
            ["bash", "-c", f"source {script_path}\n{setup_code}"], capture_output=True, text=True, timeout=120
        )
    finally:
        os.unlink(script_path)
    return result


# -- output sanity checks -----------------------------------------------------


def test_bash_completion_nontrivial(archivers, request):
    """Verify the generated Bash completion is non-trivially sized."""
    archiver = request.getfixturevalue(archivers)
    output = cmd(archiver, "completion", "bash")
    assert len(output) > 5000, f"Bash completion suspiciously small: {len(output)} chars"
    assert output.count("\n") > 100, f"Bash completion suspiciously few lines: {output.count(chr(10))}"


def test_zsh_completion_nontrivial(archivers, request):
    """Verify the generated Zsh completion is non-trivially sized."""
    archiver = request.getfixturevalue(archivers)
    output = cmd(archiver, "completion", "zsh")
    assert len(output) > 5000, f"Zsh completion suspiciously small: {len(output)} chars"
    assert output.count("\n") > 100, f"Zsh completion suspiciously few lines: {output.count(chr(10))}"


# -- syntax validation --------------------------------------------------------


def _check_shell_syntax(script_content, shell, suffix):
    """Write script_content to a temp file and verify syntax with ``shell -n``."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=suffix, delete=False) as f:
        f.write(script_content)
        script_path = f.name
    try:
        result = subprocess.run([shell, "-n", script_path], capture_output=True)
    finally:
        os.unlink(script_path)
    return result


@needs_bash
def test_bash_completion_syntax(archivers, request):
    """Verify the generated Bash completion script has valid syntax."""
    archiver = request.getfixturevalue(archivers)
    output = cmd(archiver, "completion", "bash")
    result = _check_shell_syntax(output, "bash", ".bash")
    assert result.returncode == 0, f"Generated Bash completion has syntax errors: {result.stderr.decode()}"


@needs_zsh
def test_zsh_completion_syntax(archivers, request):
    """Verify the generated Zsh completion script has valid syntax."""
    archiver = request.getfixturevalue(archivers)
    output = cmd(archiver, "completion", "zsh")
    result = _check_shell_syntax(output, "zsh", ".zsh")
    assert result.returncode == 0, f"Generated Zsh completion has syntax errors: {result.stderr.decode()}"


# -- borg-specific preamble function behavior (bash) --------------------------


@needs_bash
def test_bash_sortby_dedup(archivers, request):
    """_borg_complete_sortby should not re-offer already-selected sort keys."""
    archiver = request.getfixturevalue(archivers)
    script = cmd(archiver, "completion", "bash")

    # Simulate: user typed "borg repo-list --sort-by timestamp,"
    # The function should offer remaining keys but NOT "timestamp" again.
    result = _run_bash_completion_fn(
        script, 'COMP_WORDS=(borg repo-list --sort-by "timestamp,")\n' "COMP_CWORD=3\n" "_borg_complete_sortby\n"
    )
    assert result.returncode == 0, f"stderr: {result.stderr}"
    lines = [line for line in result.stdout.strip().splitlines() if line.strip()]
    # "timestamp" must not appear as a standalone completion candidate
    bare_keys = [line.rsplit(",", 1)[-1] for line in lines]
    assert "timestamp" not in bare_keys, f"timestamp was re-offered: {lines}"
    # Other keys like "archive" should be offered
    assert any("archive" in line for line in lines), f"expected 'archive' in completions: {lines}"


@needs_bash
def test_bash_filescachemode_exclusivity(archivers, request):
    """_borg_complete_filescachemode should enforce ctime/mtime and disabled mutual exclusion."""
    archiver = request.getfixturevalue(archivers)
    script = cmd(archiver, "completion", "bash")

    # After selecting "ctime,", mtime should not be offered
    result = _run_bash_completion_fn(
        script, 'COMP_WORDS=(borg create --files-cache "ctime,")\n' "COMP_CWORD=3\n" "_borg_complete_filescachemode\n"
    )
    assert result.returncode == 0, f"stderr: {result.stderr}"
    bare_keys = [line.rsplit(",", 1)[-1] for line in result.stdout.strip().splitlines() if line.strip()]
    assert "mtime" not in bare_keys, f"mtime offered after ctime: {bare_keys}"
    assert "disabled" not in bare_keys, f"disabled offered after ctime: {bare_keys}"

    # After selecting "disabled,", nothing should be offered
    result2 = _run_bash_completion_fn(
        script,
        'COMP_WORDS=(borg create --files-cache "disabled,")\n' "COMP_CWORD=3\n" "_borg_complete_filescachemode\n",
    )
    assert result2.returncode == 0
    assert result2.stdout.strip() == "", f"completions offered after disabled: {result2.stdout}"

    # After selecting "size,", disabled should not be offered
    result3 = _run_bash_completion_fn(
        script, 'COMP_WORDS=(borg create --files-cache "size,")\n' "COMP_CWORD=3\n" "_borg_complete_filescachemode\n"
    )
    assert result3.returncode == 0
    bare_keys3 = [line.rsplit(",", 1)[-1] for line in result3.stdout.strip().splitlines() if line.strip()]
    assert "disabled" not in bare_keys3, f"disabled offered after size: {bare_keys3}"


@needs_bash
def test_bash_archive_name_completion(archivers, request):
    """_borg_complete_archive should complete archive names from a real repo."""
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "mybackup-2024", archiver.input_path)
    cmd(archiver, "create", "mybackup-2025", archiver.input_path)

    script = cmd(archiver, "completion", "bash")
    repo = archiver.repository_path

    result = _run_bash_completion_fn(
        script, f'COMP_WORDS=(borg delete --repo "{repo}" "mybackup")\n' f"COMP_CWORD=4\n" f"_borg_complete_archive\n"
    )
    assert result.returncode == 0, f"stderr: {result.stderr}"
    assert "mybackup-2024" in result.stdout, f"archive name missing: {result.stdout}"
    assert "mybackup-2025" in result.stdout, f"archive name missing: {result.stdout}"


@needs_bash
def test_bash_archive_aid_completion(archivers, request):
    """_borg_complete_archive should complete aid: prefixed archive IDs."""
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "testarchive", archiver.input_path)

    script = cmd(archiver, "completion", "bash")
    repo = archiver.repository_path

    result = _run_bash_completion_fn(
        script, f'COMP_WORDS=(borg info --repo "{repo}" "aid:")\n' f"COMP_CWORD=4\n" f"_borg_complete_archive\n"
    )
    assert result.returncode == 0, f"stderr: {result.stderr}"
    lines = [line for line in result.stdout.strip().splitlines() if line.strip()]
    assert len(lines) >= 1, "Expected at least one archive ID completion"
    for line in lines:
        assert line.startswith("aid:"), f"Expected aid: prefix, got: {line}"



import io
import os
import sys
from collections import namedtuple
from hashlib import sha256

import pytest
from borghash import HashTableNT

from ..helpers import IntegrityError, Location, bin_to_hex
from ..hashindex import ChunkIndex
from ..repository import Repository, MAX_DATA_SIZE, rest_serve_command, PackWriter, PackReader
from ..repository import PackTracker
from ..repoobj import RepoObj, OBJ_MAGIC, OBJ_VERSION
from .hashindex_test import H


def test_rest_serve_command_local():
    # rest:// without a host runs "borg serve --rest" locally, talking over stdio.
    cmd = rest_serve_command(Location("rest:////tmp/repo"))
    assert "ssh" not in cmd
    assert cmd[0] == sys.executable
    assert cmd[-4:] == ["serve", "--rest", "--backend", "FILE:/tmp/repo"]


def test_rest_serve_command_ssh(monkeypatch):
    # rest:// with a host is reached via ssh, running "borg serve --rest" remotely.
    # we override BORGSTORE_RSH to a simple "ssh" here to simplify testing.
    # without that, borgstore 0.5.5+ would also set some ssh options via cmdline.
    monkeypatch.setenv("BORGSTORE_RSH", "ssh")
    monkeypatch.delenv("BORG_REMOTE_PATH", raising=False)
    cmd = rest_serve_command(Location("rest://user@host/repo/path"))
    assert cmd[:2] == ["ssh", "user@host"]
    assert cmd[-5:] == ["borg", "serve", "--rest", "--backend", "FILE:repo/path"]


@pytest.fixture()
def repository(tmp_path):
    repository_location = os.fspath(tmp_path / "repository")
    yield Repository(repository_location, exclusive=True, create=True)


def pytest_generate_tests(metafunc):
    # Generate tests that run on repositories.
    if "repo_fixtures" in metafunc.fixturenames:
        metafunc.parametrize("repo_fixtures", ["repository"])


def get_repository_from_fixture(repo_fixtures, request):
    # Return the repository object from the fixture.
    return request.getfixturevalue(repo_fixtures)


def reopen(repository, exclusive: bool | None = True, create=False):
    if isinstance(repository, Repository):
        if repository.opened:
            raise RuntimeError("Repo must be closed before a reopen. Cannot support nested repository contexts.")
        return Repository(repository._location, exclusive=exclusive, create=create)

    raise TypeError(f"Invalid argument type. Expected 'Repository', received '{type(repository).__name__}'.")


def fchunk(data, meta=b"", chunk_id=b"\x00" * 32):
    # Build a raw chunk with a valid RepoObj layout but no encryption or compression. Pass a unique
    # chunk_id when objects must not share a pack: identical bytes hash to the same sha256 pack id
    # and would otherwise collapse into one pack.
    hdr = RepoObj.obj_header.pack(OBJ_MAGIC, OBJ_VERSION, chunk_id, len(meta), len(data))
    assert isinstance(data, bytes)
    chunk = hdr + meta + data
    return chunk


def corrupt_chunk_on_disk(repository, chunk_id):
    # Flip a byte of the chunk's data in its pack file on disk, chunk index untouched.
    entry = repository.chunks[chunk_id]
    key = "packs/" + bin_to_hex(entry.pack_id)
    pack = repository.store_load(key)
    pos = entry.obj_offset + entry.obj_size - 1  # last byte of the chunk
    pack = pack[:pos] + bytes([pack[pos] ^ 0xFF]) + pack[pos + 1 :]
    repository.store_store(key, pack)


def pchunk(chunk):
    # Parse chunk: extract data and metadata from a raw chunk made by fchunk.
    hdr_size = RepoObj.obj_header.size
    hdr = chunk[:hdr_size]
    meta_size, data_size = RepoObj.obj_header.unpack(hdr)[3:5]
    meta = chunk[hdr_size : hdr_size + meta_size]
    data = chunk[hdr_size + meta_size : hdr_size + meta_size + data_size]
    return data, meta


def pdchunk(chunk):
    # Parse only the data from a raw chunk made by fchunk.
    return pchunk(chunk)[0]


def test_basic_operations(repo_fixtures, request):
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        for x in range(100):
            repository.put(H(x), fchunk(b"SOMEDATA", chunk_id=H(x)))  # put() updates _chunks via PackWriter
        repository.flush()  # don't rely on the put count filling whole packs; flush before get()/close()
        key50 = H(50)
        assert pdchunk(repository.get(key50)) == b"SOMEDATA"
    # no manual hand-off of the index across reopen: close() persisted it to the repo cache,
    # and the freshly opened repo rebuilds .chunks from there (or by listing the repo) on its own.
    with reopen(repository) as repository:
        for x in range(100):
            assert pdchunk(repository.get(H(x))) == b"SOMEDATA"


def test_chunk_index_persisted_on_close(tmp_path):
    # close() must serialize the live chunk index into the repo cache, so a freshly opened
    # repo can resolve pack locations without any manual hand-off. This proves the round-trip
    # by reading the persisted index back directly (not via a repo rescan, which would reconstruct
    # the same entries from the pack headers and so could mask a broken persist step).
    from ..cache import list_chunkindex_hashes, read_chunkindex_from_repo

    location = os.fspath(tmp_path / "repo")
    with Repository(location, exclusive=True, create=True) as repository:
        for x in range(10):
            repository.put(H(x), fchunk(b"DATA"))
        repository.flush()  # don't rely on the put count filling whole packs; flush before close()
    # reopen and read the cached fragments straight from disk
    with Repository(location, exclusive=True) as repository:
        persisted = ChunkIndex()
        for hash in list_chunkindex_hashes(repository):
            fragment = read_chunkindex_from_repo(repository, hash)
            if fragment is not None:
                for k, v in fragment.items():
                    persisted[k] = v
        for x in range(10):
            assert H(x) in persisted  # close() actually wrote this session's chunks
        # and the reopened repo resolves them end to end
        for x in range(10):
            assert pdchunk(repository.get(H(x))) == b"DATA"


def test_read_data(repo_fixtures, request):
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        meta, data = b"meta", b"data"
        hdr = RepoObj.obj_header.pack(OBJ_MAGIC, OBJ_VERSION, H(0), len(meta), len(data))
        chunk_complete = hdr + meta + data
        chunk_short = hdr + meta
        repository.put(H(0), chunk_complete)
        repository.flush()  # make the buffered pack durable before get()
        assert repository.get(H(0)) == chunk_complete
        assert repository.get(H(0), read_data=True) == chunk_complete
        assert repository.get(H(0), read_data=False) == chunk_short


def test_consistency(repo_fixtures, request):
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository.put(H(0), fchunk(b"foo"))
        repository.flush()  # flush before reading the just-put chunk back
        assert pdchunk(repository.get(H(0))) == b"foo"
        repository.put(H(0), fchunk(b"foo2"))
        repository.flush()
        assert pdchunk(repository.get(H(0))) == b"foo2"
        repository.put(H(0), fchunk(b"bar"))
        repository.flush()
        assert pdchunk(repository.get(H(0))) == b"bar"
        # delete removes the object the index points at; the stale earlier copies are not resurrected.
        repository.delete(H(0))
        with pytest.raises(Repository.ObjectNotFound):
            repository.get(H(0))


def test_delete_with_stale_earlier_object_in_pack(repo_fixtures, request):
    # Re-putting H(0) leaves its old bytes ahead of H(1) in the first pack while its index entry moves
    # to a new pack. delete(H(1)) then sees a partial pack view (H(1) at a non-zero offset) and must not
    # trip compact_pack's contiguity assert.
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # H(0) and H(1) share the first pack
        repository.put(H(0), fchunk(b"aaa"))
        repository.put(H(1), fchunk(b"bbb"))  # fills the pack, flushing both objects
        repository.put(H(0), fchunk(b"ccc"))  # re-put: H(0)'s index entry moves to a new pack
        repository.flush()
        repository.delete(H(1))
        with pytest.raises(Repository.ObjectNotFound):
            repository.get(H(1))
        assert pdchunk(repository.get(H(0))) == b"ccc"  # H(0) still served from its new pack


def test_multi_object_pack_roundtrip(repo_fixtures, request):
    # Two objects fill one pack and must both read back: the second from a non-zero offset, and
    # read_data=False returning only its header+meta. The test pins max_count=2 so it does not depend
    # on the Repository.open() default; the compact tests override max_count via build_one_pack() too.
    meta0, data0 = b"meta0", b"the-first-object"
    meta1, data1 = b"m1", b"second"
    chunk0 = fchunk(data0, meta=meta0, chunk_id=H(0))
    chunk1 = fchunk(data1, meta=meta1, chunk_id=H(1))
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # this test is written for exactly two objects per pack
        repository.put(H(0), chunk0)
        assert repository.chunks.is_pending(H(0))  # buffered: the pack is not full yet
        repository.put(H(1), chunk1)  # fills the pack, flushing both objects at once
        # both objects share one pack, written exactly once, laid out in put() order
        pack_id = repository.chunks[H(0)].pack_id
        assert not repository.chunks.is_pending(H(0))
        assert repository.chunks[H(1)].pack_id == pack_id
        assert [info.name for info in repository.store_list("packs")] == [bin_to_hex(pack_id)]
        assert repository.chunks[H(0)].obj_offset == 0
        assert repository.chunks[H(1)].obj_offset == len(chunk0)  # second object read from a non-zero offset
        # full reads return each object's exact bytes, the second one resolved from its non-zero offset
        assert repository.get(H(0)) == chunk0
        assert repository.get(H(1)) == chunk1
        # read_data=False returns header+meta only and stays inside the requested object
        hdr_size = RepoObj.obj_header.size
        assert repository.get(H(0), read_data=False) == chunk0[: hdr_size + len(meta0)]
        assert repository.get(H(1), read_data=False) == chunk1[: hdr_size + len(meta1)]


def test_get_many_one_load_per_pack(repo_fixtures, request):
    # get_many loads each pack once as a whole and slices every object out of the cached bytes.
    # stats["load_calls"] counts store.load() calls.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(6)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 3  # three objects per pack -> two packs for the six objects
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        ids = [chunk_id for chunk_id, _ in objects]
        assert len({repository.chunks[chunk_id].pack_id for chunk_id in ids}) == 2  # six ids, two packs
        one_by_one = [repository.get(chunk_id) for chunk_id in ids]  # reference bytes, one store.load each

        loads_before = repository.store.stats["load_calls"]
        assert list(repository.get_many(ids)) == one_by_one
        assert repository.store.stats["load_calls"] - loads_before == 2  # one store.load per pack


def test_get_many_keeps_request_order(repo_fixtures, request):
    # Ids requested out of their stored order come back in the requested order with their own bytes.
    # Interleaving two packs still loads each pack once: the pack cache serves the second visit.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(6)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 3  # two packs: {H0,H1,H2} and {H3,H4,H5}
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        ids = [H(0), H(3), H(1), H(4), H(2), H(5)]  # interleave the two packs
        one_by_one = [repository.get(chunk_id) for chunk_id in ids]

        loads_before = repository.store.stats["load_calls"]
        assert list(repository.get_many(ids)) == one_by_one  # same order, same bytes
        assert repository.store.stats["load_calls"] - loads_before == 2  # each pack loaded once


def test_get_many_missing_id_yields_none(repo_fixtures, request):
    # With raise_missing=False, an id that was never stored yields None in its position; the ids
    # before and after it read back unchanged.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(3)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 4  # above the object count, so the pack flushes on flush() only
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        repository.flush()
        ids = [H(0), H(99), H(2)]  # H(99) was never put
        result = list(repository.get_many(ids, raise_missing=False))
        assert result[0] == repository.get(H(0))
        assert result[1] is None  # never stored -> None
        assert result[2] == repository.get(H(2))


def test_get_many_missing_pack_raises_pack_not_found(repo_fixtures, request):
    # get_many raises PackNotFound when a chunk's index entry points to a pack that is not in the
    # store; with raise_missing=False it yields None instead.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(2)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # one pack: {H0, H1}
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        repository.flush()
        pack_id = repository.chunks[H(0)].pack_id
        repository.store_delete("packs/" + bin_to_hex(pack_id))  # delete the pack, keep its index entry

        with pytest.raises(Repository.PackNotFound):
            list(repository.get_many([H(0)]))
        with pytest.raises(Repository.PackNotFound):
            list(repository.get_many([H(0), H(1)]))  # one id's pack missing in a batch
        assert list(repository.get_many([H(0)], raise_missing=False)) == [None]  # raise_missing=False -> None


def test_get_many_repeated_ids(repo_fixtures, request):
    # A dedup'd item repeats chunk ids, e.g. [A, A, B, C, B]. Each repeat returns the right bytes
    # and each pack is loaded once: the cached pack serves the later visits.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(4)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # two packs: {H0,H1} and {H2,H3}
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        assert len({repository.chunks[H(i)].pack_id for i in range(4)}) == 2
        ids = [H(0), H(0), H(1), H(2), H(1)]  # A A B C B, A/B in one pack, C in the other
        one_by_one = [repository.get(chunk_id) for chunk_id in ids]

        loads_before = repository.store.stats["load_calls"]
        assert list(repository.get_many(ids)) == one_by_one  # every position, including repeats
        assert repository.store.stats["load_calls"] - loads_before == 2  # two packs, one load each


def test_get_many_evicts_least_recently_used(repo_fixtures, request):
    # Visiting more packs than the cache holds evicts the oldest; revisiting an evicted pack reloads it.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(8)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # four packs: {H0,H1} {H2,H3} {H4,H5} {H6,H7}
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        assert len({repository.chunks[H(i)].pack_id for i in range(8)}) == 4
        assert repository.PACK_READER_CACHE_SIZE == 3  # cache holds three of the four packs
        # touch packs 0,1,2 (fills cache), then 3 (evicts pack 0), then pack 0 again (reload).
        ids = [H(0), H(2), H(4), H(6), H(0)]
        one_by_one = [repository.get(chunk_id) for chunk_id in ids]

        loads_before = repository.store.stats["load_calls"]
        assert list(repository.get_many(ids)) == one_by_one
        assert repository.store.stats["load_calls"] - loads_before == 5  # 4 distinct packs + 1 reload


def test_get_reuses_cached_pack(repo_fixtures, request):
    # get() slices from a pack that get_many already cached, doing no store load; a get() whose pack
    # is not cached reads only its object's range and does not load the whole pack into the cache.
    objects = [(H(i), fchunk(b"payload-%02d" % i, chunk_id=H(i))) for i in range(2)]
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        repository._pack_writer.max_count = 2  # one pack: {H0, H1}
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)

        # cold cache: get() reads one range and leaves the cache empty, so the next get() reads again
        loads_before = repository.store.stats["load_calls"]
        reference = repository.get(H(0))
        assert repository.store.stats["load_calls"] - loads_before == 1  # one ranged read
        loads_before = repository.store.stats["load_calls"]
        repository.get(H(1))
        assert repository.store.stats["load_calls"] - loads_before == 1  # pack was not cached, another range

        list(repository.get_many([H(0), H(1)]))  # loads the whole pack into the cache

        loads_before = repository.store.stats["load_calls"]
        assert repository.get(H(0)) == reference
        assert repository.store.stats["load_calls"] - loads_before == 0  # sliced from the cached pack
        assert repository.get(H(1), read_data=False)  # read_data=False also peeks the cache
        assert repository.store.stats["load_calls"] - loads_before == 0


def build_one_pack(repository, objects):
    with repository:
        repository._pack_writer.max_count = len(objects) + 1  # prevent per-put flush; one pack on flush()
        for chunk_id, chunk in objects:
            repository.put(chunk_id, chunk)
        repository.flush()


def test_compact_pack_copy_forward(repo_fixtures, request):
    # Keep a subset of a multi-object pack: kept objects must read back, the dropped object and its bytes gone.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    chunk2 = fchunk(b"DATA2", chunk_id=H(2))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1), (H(2), chunk2)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        assert repository.chunks[H(1)].pack_id == old_pack_id
        assert repository.chunks[H(2)].pack_id == old_pack_id

        new_pack_id, dropped = repository.compact_pack(old_pack_id, keep_ids={H(0), H(2)}, drop_ids={H(1)})

        assert new_pack_id is not None and new_pack_id != old_pack_id
        assert dropped == len(chunk1)  # reported freed bytes for --stats
        assert pdchunk(repository.get(H(0))) == b"DATA0"
        assert pdchunk(repository.get(H(2))) == b"DATA2"
        assert repository.get(H(1), raise_missing=False) is None  # compact_pack removed its index entry
        packs = {info.name: info.size for info in repository.store_list("packs")}
        assert bin_to_hex(old_pack_id) not in packs
        assert packs[bin_to_hex(new_pack_id)] == len(chunk0) + len(chunk2)  # only the kept objects' bytes


def test_compact_pack_drops_whole_pack(repo_fixtures, request):
    # Dropping every object removes the pack and clears its index entries.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id

        new_pack_id, dropped = repository.compact_pack(old_pack_id, keep_ids=set(), drop_ids={H(0), H(1)})
        assert new_pack_id is None  # every byte dropped: no replacement pack
        assert dropped == len(chunk0) + len(chunk1)

        assert repository.get(H(0), raise_missing=False) is None
        assert repository.get(H(1), raise_missing=False) is None
        assert bin_to_hex(old_pack_id) not in [info.name for info in repository.store_list("packs")]


def test_compact_pack_keep_all_is_noop(repo_fixtures, request):
    # Keeping every object reproduces the same pack: same sha256 name, old pack not deleted. Ids passed
    # out of order must give the same result, since compact_pack sorts by offset.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id

        new_pack_id, dropped = repository.compact_pack(
            old_pack_id, keep_ids={H(1), H(0)}, drop_ids=set()
        )  # out of order

        assert new_pack_id == old_pack_id
        assert dropped == 0  # nothing dropped
        assert pdchunk(repository.get(H(0))) == b"DATA0"
        assert pdchunk(repository.get(H(1))) == b"DATA1"


def test_compact_pack_keeps_gap(repo_fixtures, request):
    # A pack may hold bytes no index entry covers (a superseded copy of a chunk re-put elsewhere).
    # compact_pack must copy such gaps into the new pack rather than dropping data it was not told to
    # drop; only "borg check --repair" decides their fate (issue #9868). Drop the middle object's index
    # entry to make a gap, then rewrite dropping the first object and keeping the last: the gap survives.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    chunk2 = fchunk(b"DATA2", chunk_id=H(2))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1), (H(2), chunk2)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        del repository.chunks[H(1)]  # H(1)'s bytes stay in the pack but are now unindexed (a gap)

        new_pack_id, _ = repository.compact_pack(old_pack_id, keep_ids={H(2)}, drop_ids={H(0)})

        assert new_pack_id is not None and new_pack_id != old_pack_id
        assert pdchunk(repository.get(H(2))) == b"DATA2"
        packs = {info.name: info.size for info in repository.store_list("packs")}
        assert bin_to_hex(old_pack_id) not in packs
        assert packs[bin_to_hex(new_pack_id)] == len(chunk1) + len(chunk2)  # gap bytes kept, only H(0) dropped


def test_compact_pack_keeps_trailing_bytes(repo_fixtures, request):
    # Same as the gap case, but the unindexed bytes are at the end of the pack: dropping the last
    # object's index entry leaves trailing bytes the listed objects do not reach. Rewrite dropping the
    # first object; the trailing bytes must be copied forward, not lost.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    chunk2 = fchunk(b"DATA2", chunk_id=H(2))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1), (H(2), chunk2)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        del repository.chunks[H(2)]  # trailing unindexed bytes

        new_pack_id, _ = repository.compact_pack(old_pack_id, keep_ids={H(1)}, drop_ids={H(0)})

        assert new_pack_id is not None and new_pack_id != old_pack_id
        assert pdchunk(repository.get(H(1))) == b"DATA1"
        packs = {info.name: info.size for info in repository.store_list("packs")}
        assert packs[bin_to_hex(new_pack_id)] == len(chunk1) + len(chunk2)  # trailing bytes kept


def test_compact_pack_drops_superseded_gap(repo_fixtures, request):
    # A gap object whose chunk id is still in the index has its authoritative copy elsewhere, so it is
    # a superseded duplicate and compact_pack drops its bytes. Repoint H(1) to another pack: its bytes
    # here become an indexed-elsewhere gap. Rewrite keeping H(0) and H(2); the gap is dropped and H(1)'s
    # index entry (pointing elsewhere) stays.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    chunk2 = fchunk(b"DATA2", chunk_id=H(2))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1), (H(2), chunk2)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        repository.chunks[H(1)] = repository.chunks[H(1)]._replace(pack_id=H(9))  # authoritative copy elsewhere

        new_pack_id, dropped = repository.compact_pack(old_pack_id, keep_ids={H(0), H(2)}, drop_ids=set())

        assert new_pack_id is not None and new_pack_id != old_pack_id
        assert dropped == len(chunk1)  # the superseded gap's bytes are counted as freed
        assert pdchunk(repository.get(H(0))) == b"DATA0"
        assert pdchunk(repository.get(H(2))) == b"DATA2"
        packs = {info.name: info.size for info in repository.store_list("packs")}
        assert bin_to_hex(old_pack_id) not in packs
        assert packs[bin_to_hex(new_pack_id)] == len(chunk0) + len(chunk2)  # superseded H(1) gap dropped
        assert repository.chunks[H(1)].pack_id == H(9)  # index entry pointing elsewhere is untouched


def test_compact_pack_keeps_self_referencing_gap(repo_fixtures, request):
    # A gap object whose index entry points back at this very pack and offset is the object's only copy
    # (a caller that broke the keep+drop contract by not listing it). compact_pack must keep its bytes,
    # not destroy them. Leave H(1) out of keep_ids and drop_ids with its entry pointing here.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    chunk2 = fchunk(b"DATA2", chunk_id=H(2))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1), (H(2), chunk2)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id

        new_pack_id, dropped = repository.compact_pack(old_pack_id, keep_ids={H(0), H(2)}, drop_ids=set())

        assert new_pack_id == old_pack_id  # nothing dropped, defrag reproduced the same pack
        assert dropped == 0  # the self-referencing gap is kept, nothing freed
        packs = {info.name: info.size for info in repository.store_list("packs")}
        assert packs[bin_to_hex(new_pack_id)] == len(chunk0) + len(chunk1) + len(chunk2)  # H(1) bytes kept


def test_compact_pack_detects_overlap(repo_fixtures, request):
    # Overlapping index entries mean index corruption: compact_pack must raise IntegrityError and
    # leave the pack in place. Point H(1) back at H(0)'s offset to overlap it.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        entry = repository.chunks[H(1)]
        repository.chunks[H(1)] = entry._replace(obj_offset=0)  # now overlaps H(0) at offset 0

        with pytest.raises(IntegrityError):
            repository.compact_pack(old_pack_id, keep_ids={H(0), H(1)}, drop_ids=set())
        assert bin_to_hex(old_pack_id) in [info.name for info in repository.store_list("packs")]


def test_compact_pack_detects_past_eof(repo_fixtures, request):
    # An index entry whose span ends past the pack file means index corruption: compact_pack must
    # raise IntegrityError and leave the pack in place, so defrag never short-reads a truncated object.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id
        entry = repository.chunks[H(1)]
        repository.chunks[H(1)] = entry._replace(obj_size=entry.obj_size + 1000)  # now claims to end past EOF

        with pytest.raises(IntegrityError):
            repository.compact_pack(old_pack_id, keep_ids={H(0), H(1)}, drop_ids=set())
        assert bin_to_hex(old_pack_id) in [info.name for info in repository.store_list("packs")]


def test_compact_pack_translates_read_range_error(repo_fixtures, request, monkeypatch):
    # On a truncated or corrupt pack a defrag span reads back short and borgstore raises ReadRangeError.
    # compact_pack translates it to IntegrityError, keeps the source pack and leaves the index unchanged.
    # store.defrag is patched to raise ReadRangeError.
    from borgstore.store import ReadRangeError

    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        old_pack_id = repository.chunks[H(0)].pack_id

        def short_read(*args, **kwargs):
            raise ReadRangeError("requested 5 bytes at offset 0, got 3")

        monkeypatch.setattr(repository.store, "defrag", short_read)
        with pytest.raises(IntegrityError):
            repository.compact_pack(old_pack_id, keep_ids={H(0)}, drop_ids={H(1)})
        assert bin_to_hex(old_pack_id) in [info.name for info in repository.store_list("packs")]
        assert H(1) in repository.chunks  # still indexed: aborted before deleting the dropped id


def test_merge_packs_combines_whole_files(repo_fixtures, request):
    # Several one-object packs merge into one, and every object reads back from its new location.
    repository = get_repository_from_fixture(repo_fixtures, request)
    with repository:
        pack_ids = []
        for i in range(3):
            repository.put(H(i), fchunk(f"DATA{i}".encode(), chunk_id=H(i)))
            repository.flush()  # one object per pack
            pack_ids.append(repository.chunks[H(i)].pack_id)

        packs_before = {info.name for info in repository.store_list("packs")}
        assert len(packs_before) == 3

        repository.merge_packs(pack_ids)

        packs_after = {info.name for info in repository.store_list("packs")}
        assert len(packs_after) == 1
        assert packs_after.isdisjoint(packs_before)  # a brand-new pack, all originals deleted
        for i in range(3):
            assert pdchunk(repository.get(H(i))) == f"DATA{i}".encode()
            assert bin_to_hex(repository.chunks[H(i)].pack_id) in packs_after


def test_merge_packs_carries_unindexed_bytes_forward(repo_fixtures, request):
    # A pack byte range no index entry covers (e.g. a chunk copy superseded by a later put) must
    # survive a merge unchanged, not be silently dropped: merge_packs copies whole pack files.
    # Merged alongside a second pack, so the result is a genuinely new pack (a lone unchanged pack
    # would defrag to identical bytes and thus the same content-addressed name).
    chunk0 = fchunk(b"AAAA", chunk_id=H(0))
    chunk1 = fchunk(b"BBBB", chunk_id=H(1))  # this entry is removed from the index before merging
    chunk2 = fchunk(b"CCCC", chunk_id=H(2))  # lives in a second, separate pack
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0), (H(1), chunk1)])
    with repository:
        repository.put(H(2), chunk2)
        repository.flush()
        pack1_id = repository.chunks[H(0)].pack_id
        pack2_id = repository.chunks[H(2)].pack_id
        pack1_size_before = repository.store.info("packs/" + bin_to_hex(pack1_id)).size
        del repository.chunks[H(1)]  # H(1)'s bytes are now unindexed, but still on disk

        repository.merge_packs({pack1_id, pack2_id})

        new_pack_id = repository.chunks[H(0)].pack_id
        assert new_pack_id == repository.chunks[H(2)].pack_id  # both merged into the same new pack
        new_pack_size = repository.store.info("packs/" + bin_to_hex(new_pack_id)).size
        assert new_pack_size == pack1_size_before + len(chunk2)  # H(1)'s bytes carried forward, not dropped
        assert pdchunk(repository.get(H(0))) == b"AAAA"
        assert pdchunk(repository.get(H(2))) == b"CCCC"


def test_merge_packs_detects_past_eof(repo_fixtures, request):
    # An index entry claiming bytes past its pack's actual end means index corruption. merge_packs
    # must raise before writing anything, so the only intact copy of the pack is never deleted.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    repository = get_repository_from_fixture(repo_fixtures, request)
    build_one_pack(repository, [(H(0), chunk0)])
    with repository:
        pack_id = repository.chunks[H(0)].pack_id
        entry = repository.chunks[H(0)]
        repository.chunks[H(0)] = entry._replace(obj_size=entry.obj_size + 1)  # claims 1 byte past EOF

        with pytest.raises(IntegrityError):
            repository.merge_packs({pack_id})
        assert bin_to_hex(pack_id) in [info.name for info in repository.store_list("packs")]  # untouched


def test_merge_packs_skips_pack_already_gone(repo_fixtures, request):
    # A stale index entry pointing at a pack file the store no longer has (#9850) must not abort the
    # merge: the pack is excluded with a warning and the rest of the batch still merges.
    chunk0 = fchunk(b"DATA0", chunk_id=H(0))
    chunk1 = fchunk(b"DATA1", chunk_id=H(1))
    repository = get_repository_from_fixture(repo_fixtures, request)
    with repository:
        repository.put(H(0), chunk0)
        repository.flush()
        pack0_id = repository.chunks[H(0)].pack_id
        repository.put(H(1), chunk1)
        repository.flush()
        pack1_id = repository.chunks[H(1)].pack_id

        repository.store_delete("packs/" + bin_to_hex(pack0_id))  # pack gone, index entry left stale

        repository.merge_packs({pack0_id, pack1_id})

        assert pdchunk(repository.get(H(1))) == b"DATA1"  # the still-present pack merged normally


def test_assert_writable(repository):
    # Compaction needs write (w/W) and delete (D) on both packs/ and index/. assert_writable reads
    # self.permissions directly, so set it explicitly to cover each case.
    with repository:
        repository.permissions = None  # "all": no restrictions
        repository.assert_writable()  # does not raise

        repository.permissions = {"packs": "lrwWD", "index": "lrwWD"}
        repository.assert_writable()  # does not raise

        repository.permissions = {"packs": "lrw", "index": "lrwWD"}  # packs/ has no delete
        with pytest.raises(Repository.PermissionDenied):
            repository.assert_writable()

        repository.permissions = {"packs": "lrwWD", "index": "lrD"}  # index/ has no write
        with pytest.raises(Repository.PermissionDenied):
            repository.assert_writable()

        repository.permissions = {"": "lr"}  # neither namespace listed; "" fallback grants read only
        with pytest.raises(Repository.PermissionDenied):
            repository.assert_writable()


def test_list(repo_fixtures, request):
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        for x in range(100):
            repository.put(H(x), fchunk(b"SOMEDATA", chunk_id=H(x)))  # unique bytes -> unique pack id
        repository.flush()  # flush the last partial pack so all 100 objects are listable
        repo_list = repository.list()
        assert len(repo_list) == 100
        first_half = repository.list(limit=50)
        assert len(first_half) == 50
        assert first_half == repo_list[:50]
        second_half = repository.list(marker=first_half[-1][0])
        assert len(second_half) == 50
        assert second_half == repo_list[50:]
        assert len(repository.list(limit=50)) == 50


def test_max_data_size(repo_fixtures, request):
    with get_repository_from_fixture(repo_fixtures, request) as repository:
        max_data = b"x" * (MAX_DATA_SIZE - RepoObj.obj_header.size)
        repository.put(H(0), fchunk(max_data))
        repository.flush()  # make the buffered pack durable before get()
        assert pdchunk(repository.get(H(0))) == max_data
        with pytest.raises(IntegrityError):
            repository.put(H(1), fchunk(max_data + b"x"))
        repository.delete(H(0))


def check(repository, repo_path, repair=False, status=True):
    assert repository.check(repair=repair) == status
    # Make sure no tmp files are left behind
    tmp_files = [name for name in os.listdir(repo_path) if "tmp" in name]
    assert tmp_files == [], "Found tmp files"


class MockStore:
    def __init__(self):
        self.stored = {}

    def store(self, key, data):
        self.stored[key] = data


class FailingPackStore:
    """Wraps a store but fails packs/* writes; every other call passes through to the inner store.

    Models the realistic failure where only a pack write broke while the rest of the repo (e.g. the
    index/* objects) stay writable.  In production PackWriter and the chunk index share
    one store, so a single object has to fail the pack write yet still let the index persist.
    """

    def __init__(self, inner):
        self._inner = inner

    def store(self, key, data):
        if key.startswith("packs/"):
            raise OSError("simulated pack store failure")
        return self._inner.store(key, data)

    def __getattr__(self, name):
        return getattr(self._inner, name)


def test_pack_writer_returns_none_when_not_full():
    pw = PackWriter(MockStore(), max_count=2, chunks=ChunkIndex())
    assert pw.add(b"a" * 32, b"data") is None


def test_pack_writer_flush_returns_none_when_empty():
    pw = PackWriter(MockStore(), max_count=1, chunks=ChunkIndex())
    assert pw.flush() is None


def test_pack_writer_n1_flush():
    store = MockStore()
    chunk_id = b"c" * 32
    cdata = b"payload"
    pw = PackWriter(store, max_count=1, chunks=ChunkIndex())
    results = pw.add(chunk_id, cdata)
    assert results is not None
    assert len(results) == 1
    stored_id, pack_id, obj_offset, obj_size = results[0]
    assert stored_id == chunk_id
    assert pack_id == sha256(cdata).digest()
    assert obj_offset == 0
    assert obj_size == len(cdata)


def test_pack_writer_n2_flush():
    store = MockStore()
    id1, id2 = b"a" * 32, b"b" * 32
    data1, data2 = b"first", b"second"
    pw = PackWriter(store, max_count=2, chunks=ChunkIndex())
    assert pw.add(id1, data1) is None
    results = pw.add(id2, data2)
    assert results is not None
    assert len(results) == 2
    pack_data = data1 + data2
    expected_pack_id = sha256(pack_data).digest()
    assert results[0] == (id1, expected_pack_id, 0, len(data1))
    assert results[1] == (id2, expected_pack_id, len(data1), len(data2))


def test_pack_writer_flushes_on_max_size():
    # max_count is high, so the flush is driven by max_size alone.
    store = MockStore()
    pw = PackWriter(store, max_count=100, max_size=10, chunks=ChunkIndex())
    assert pw.add(b"a" * 32, b"12345") is None
    results = pw.add(b"b" * 32, b"67890")
    assert results is not None
    assert len(results) == 2


def test_pack_writer_max_size_none_is_count_only():
    store = MockStore()
    pw = PackWriter(store, max_count=2, max_size=None, chunks=ChunkIndex())
    assert pw.add(b"a" * 32, b"x" * 10_000) is None
    assert pw.add(b"b" * 32, b"y" * 10_000) is not None


def test_pack_writer_max_count_none_is_size_only():
    store = MockStore()
    pw = PackWriter(store, max_count=None, max_size=10, chunks=ChunkIndex())
    assert pw.add(b"a" * 32, b"12345") is None
    assert pw.add(b"b" * 32, b"67890") is not None


def test_pack_writer_requires_a_limit():
    with pytest.raises(ValueError):
        PackWriter(MockStore(), max_count=None, max_size=None, chunks=ChunkIndex())


def test_pack_writer_rolls_back_index_on_failed_store():
    # If store.store() fails, flush() must drop the entries add() pre-marked, otherwise the index
    # keeps a phantom (indexed but never stored) chunk that seen_chunk() reports as present and a
    # later identical chunk would dedup against -- silent data loss (#9744 review).
    chunks = ChunkIndex()
    chunk_id = b"e" * 32
    pw = PackWriter(FailingPackStore(MockStore()), max_count=1, chunks=chunks)
    with pytest.raises(OSError):
        pw.add(chunk_id, b"payload")  # max_count=1 -> add() flushes immediately and fails
    assert chunks.get(chunk_id) is None  # rolled back: no phantom entry left behind


def test_failed_store_phantom_not_persisted(tmp_path):
    # The phantom must not survive into the persisted repo index either: close() can write the
    # in-memory index on context exit, so the rollback has to happen before anything is serialized.
    from ..cache import write_chunkindex_to_repo, build_chunkindex_from_repo

    chunk_id = H(60)
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        # fail only the pack write on the repository's own store; index/* writes still work,
        # so one store models "just the pack write broke" (PackWriter and the index share a
        # store in production). the failing store is thus load-bearing for every assertion below.
        repository.store = FailingPackStore(repository.store)
        pw = PackWriter(repository.store, max_count=1, repository=repository)
        with pytest.raises(OSError):
            pw.add(chunk_id, fchunk(b"DATA"))
        assert repository.chunks.get(chunk_id) is None  # rolled back from the in-memory index ...
        # ... and persisting + reloading the index (through that same store) does not bring it back:
        write_chunkindex_to_repo(repository, repository.chunks, incremental=True)
        reloaded = build_chunkindex_from_repo(repository)
        assert reloaded.get(chunk_id) is None


def test_get_read_data_false_with_range(tmp_path):
    # read_data=False with ChunkIndex entries limits the load to each object's boundary.
    hdr_size = RepoObj.obj_header.size
    chunk1 = fchunk(b"FIRST")
    chunk2 = fchunk(b"SECOND")
    pack = chunk1 + chunk2
    pack_id = H(43)
    id1, id2 = H(47), H(48)
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store("packs/" + bin_to_hex(pack_id), pack)
        chunks = ChunkIndex()
        chunks.add(id1, len(chunk1))
        chunks.update_pack_info([(id1, pack_id, 0, len(chunk1))])
        chunks.add(id2, len(chunk2))
        chunks.update_pack_info([(id2, pack_id, len(chunk1), len(chunk2))])
        repository.chunks = chunks
        assert repository.get(id1, read_data=False) == chunk1[:hdr_size]
        assert repository.get(id2, read_data=False) == chunk2[:hdr_size]


def test_get_read_data_false_large_meta(tmp_path):
    # When meta_size > extra_size (975 bytes), get() retries with a larger load.
    hdr_size = RepoObj.obj_header.size
    # the first try loads ~1KB, so use a meta clearly past that boundary to force the retry path.
    big_meta = b"M" * 5000
    chunk = fchunk(b"DATA", meta=big_meta)
    pack_id = H(44)
    chunk_id = H(49)
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store("packs/" + bin_to_hex(pack_id), chunk)
        chunks = ChunkIndex()
        chunks.add(chunk_id, len(chunk))
        chunks.update_pack_info([(chunk_id, pack_id, 0, len(chunk))])
        repository.chunks = chunks
        result = repository.get(chunk_id, read_data=False)
        assert result == chunk[: hdr_size + len(big_meta)]


def test_get_uses_chunk_index_location(tmp_path):
    # get() routes to the correct pack and offset when a ChunkIndex is assigned via the chunks property.
    chunk1 = fchunk(b"FIRST")
    chunk2 = fchunk(b"SECOND")
    pack = chunk1 + chunk2
    pack_id = H(55)
    id1, id2 = H(56), H(57)
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        # Inject the pack directly; bypasses PackWriter to test routing independently.
        repository.store_store("packs/" + bin_to_hex(pack_id), pack)
        chunks = ChunkIndex()
        chunks.add(id1, len(chunk1))
        chunks.update_pack_info([(id1, pack_id, 0, len(chunk1))])
        chunks.add(id2, len(chunk2))
        chunks.update_pack_info([(id2, pack_id, len(chunk1), len(chunk2))])
        repository.chunks = chunks
        assert repository.get(id1) == chunk1
        assert repository.get(id2) == chunk2


def test_put_marks_id_in_chunk_index(tmp_path):
    # put() marks the id pending; flush() sets the real pack location and clears the pending flag.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        id1 = H(1)
        repository.put(id1, fchunk(b"ZEROS"))
        entry = repository._chunks.get(id1)
        assert entry is not None
        assert repository._chunks.is_pending(id1)  # buffered, not yet flushed
        repository.flush()
        entry = repository._chunks.get(id1)
        assert not repository._chunks.is_pending(id1)
        assert entry.pack_id == sha256(fchunk(b"ZEROS")).digest()
        assert entry.size == 0  # uncompressed size filled in by cache layer


def test_list_skips_pending_chunk(tmp_path):
    # list() skips a pending chunk and yields it once flushed.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.put(H(1), fchunk(b"BUFFERED"))  # buffered: the pack is not full yet
        assert repository._chunks.is_pending(H(1))
        assert repository.list() == []
        repository.flush()
        assert [chunk_id for chunk_id, _ in repository.list()] == [H(1)]


def test_get_pending_chunk_raises(tmp_path):
    # get() on a pending chunk raises PackLocationUnknown, also with raise_missing=False.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.put(H(1), fchunk(b"BUFFERED"))  # buffered: the pack is not full yet
        assert repository._chunks.is_pending(H(1))
        with pytest.raises(Repository.PackLocationUnknown):
            repository.get(H(1))
        with pytest.raises(Repository.PackLocationUnknown):
            repository.get(H(1), raise_missing=False)
        repository.flush()  # close() requires the buffer to be empty


def test_flush_store_failure_drops_pending_entries(tmp_path):
    # flush() removes the pending index entries when storing the pack fails.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.put(H(1), fchunk(b"BUFFERED"))
        assert repository._chunks.is_pending(H(1))

        def boom(*args, **kwargs):
            raise OSError("store failed")

        repository._pack_writer.store.store = boom
        with pytest.raises(OSError):
            repository.flush()
        assert H(1) not in repository._chunks


def test_check_detects_corruption_in_later_object(tmp_path):
    # Corruption anywhere in a multi-object pack must be caught, not just in the first object: the pack
    # is named by sha256(content), so flipping any byte makes its stored hash differ from its name.
    chunk1 = fchunk(b"FIRST", chunk_id=H(1))
    chunk2 = fchunk(b"SECOND", chunk_id=H(2))
    pack = chunk1 + chunk2
    pack_name = "packs/" + bin_to_hex(sha256(pack).digest())
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store(pack_name, pack)
        assert repository.check(repair=False) is True  # both objects are intact

        # flip a byte of the SECOND object's OBJ_MAGIC; the first object stays valid.
        corrupted = bytearray(pack)
        corrupted[len(chunk1)] ^= 0xFF
        repository.store_store(pack_name, bytes(corrupted))
        assert repository.check(repair=False) is False  # corruption past object 1 is detected


def test_check_detects_index_corruption(tmp_path):
    # index/ objects are named by sha256(content) like packs, so check verifies them the same way.
    content = b"pretend this is a serialized chunk index"
    index_name = "index/" + bin_to_hex(sha256(content).digest())
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store(index_name, content)
        assert repository.check(repair=False) is True  # index object intact (name == sha256(content))

        corrupted = bytearray(content)
        corrupted[0] ^= 0xFF
        repository.store_store(index_name, bytes(corrupted))  # same name, rotted content
        assert repository.check(repair=False) is False  # mismatch between content hash and name detected


def test_check_warns_on_invalid_chunk_index(tmp_path, caplog):
    # check warns about an invalid chunk index but does not fail, since the index is not part of
    # the repository's object integrity.
    import logging
    from ..cache import write_chunkindex_invalid

    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        write_chunkindex_invalid(repository)
        with caplog.at_level(logging.WARNING):
            assert repository.check(repair=False) is True
        assert "chunk index is invalid" in caplog.text


def test_check_intact_multi_object_pack_passes(tmp_path):
    # An intact pack with several objects passes: it is hashed as a whole, so the object count
    # does not matter.
    pack = fchunk(b"A", chunk_id=H(1)) + fchunk(b"BB", chunk_id=H(2)) + fchunk(b"CCC", chunk_id=H(3))
    pack_name = "packs/" + bin_to_hex(sha256(pack).digest())
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store(pack_name, pack)
        assert repository.check(repair=False) is True


def test_check_checked_packs_roundtrip(tmp_path):
    # the set survives a store/load round-trip; a rotted blob loads as empty.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        tracker = PackTracker.new(repository.store)
        tracker.table[H(1)] = PackTracker.Entry(timestamp=123, result=1)
        tracker.table[H(2)] = PackTracker.Entry(timestamp=456, result=0)
        tracker.save()

        loaded = PackTracker.load(repository.store)
        assert len(loaded) == 2
        assert H(1) in loaded.table and H(2) in loaded.table
        assert tuple(loaded.table[H(2)]) == (456, 0)

        corrupted = bytearray(repository.store.load(PackTracker.NAME))
        corrupted[0] ^= 0xFF  # break the appended sha256
        repository.store.store(PackTracker.NAME, bytes(corrupted))
        rotted = PackTracker.load(repository.store)
        assert len(rotted) == 0


def test_check_partial_rechecks_pack_sorting_before_checked_one(tmp_path):
    # a partial check verifies a new pack even when its id sorts before an already-checked pack.
    intact = fchunk(b"INTACT", chunk_id=H(1))
    intact_id = sha256(intact).digest()
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store("packs/" + bin_to_hex(intact_id), intact)

        # mark the intact pack as already checked in this cycle.
        tracker = PackTracker.new(repository.store)
        tracker.record(intact_id, ok=True)
        tracker.save()

        # add a corrupt pack (content does not hash to its name) whose id sorts before intact_id.
        early_id = b"\x00" * 32
        assert bin_to_hex(early_id) < bin_to_hex(intact_id)
        repository.store_store("packs/" + bin_to_hex(early_id), b"CORRUPT-does-not-match-name")

        assert repository.check(repair=False, max_duration=3600) is False


def test_check_partial_rechecks_pack_recorded_corrupt(tmp_path):
    # a pack recorded corrupt earlier in the cycle is re-verified, so the corruption keeps being reported.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        corrupt_id = H(1)  # stored content does not hash to this name
        repository.store_store("packs/" + bin_to_hex(corrupt_id), b"CORRUPT-does-not-match-name")

        tracker = PackTracker.new(repository.store)
        tracker.record(corrupt_id, ok=False)
        tracker.save()

        assert repository.check(repair=False, max_duration=3600) is False


def _spy_hash(repository, monkeypatch):
    # collect the keys passed to store.hash. check() hashes a pack exactly when it verifies it.
    hashed_keys = []
    orig_hash = repository.store.hash

    def spy_hash(key):
        hashed_keys.append(key)
        return orig_hash(key)

    monkeypatch.setattr(repository.store, "hash", spy_hash)
    return hashed_keys


def _store_intact_pack(repository):
    intact = fchunk(b"INTACT", chunk_id=H(1))
    intact_id = sha256(intact).digest()
    pack_key = "packs/" + bin_to_hex(intact_id)
    repository.store_store(pack_key, intact)
    return intact_id, pack_key


def test_check_partial_clears_recorded_corruption_when_intact(tmp_path, monkeypatch):
    # a pack recorded corrupt is re-verified, and an intact result replaces the stale record.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        intact_id, pack_key = _store_intact_pack(repository)

        tracker = PackTracker.new(repository.store)
        tracker.record(intact_id, ok=False)  # stale corrupt record
        tracker.save()

        hashed_keys = _spy_hash(repository, monkeypatch)

        assert repository.check(repair=False, max_duration=3600) is True
        assert pack_key in hashed_keys  # re-verified


def test_check_partial_skips_pack_recorded_intact(tmp_path, monkeypatch):
    # a pack recorded intact in this cycle is skipped when a partial check resumes.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        intact_id, pack_key = _store_intact_pack(repository)

        tracker = PackTracker.new(repository.store)
        tracker.record(intact_id, ok=True)
        tracker.save()

        hashed_keys = _spy_hash(repository, monkeypatch)

        assert repository.check(repair=False, max_duration=3600) is True
        assert pack_key not in hashed_keys  # skipped, not re-verified


def test_check_full_ignores_recorded_set(tmp_path, monkeypatch):
    # a full check verifies every pack regardless of the recorded set, then drops the set.
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        intact_id, pack_key = _store_intact_pack(repository)

        tracker = PackTracker.new(repository.store)
        tracker.record(intact_id, ok=True)
        tracker.save()

        hashed_keys = _spy_hash(repository, monkeypatch)

        assert repository.check(repair=False) is True
        assert pack_key in hashed_keys  # verified

        after = PackTracker.load(repository.store)
        assert len(after) == 0  # cycle complete, set dropped


def test_check_checked_packs_ignores_foreign_entry_layout(tmp_path):
    # load() drops a set whose entries have a different layout than Entry, even though its sha256 matches.
    OtherEntry = namedtuple("OtherEntry", "timestamp result extra")
    OtherFormat = namedtuple("OtherFormat", "timestamp result extra")
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        table = HashTableNT(
            key_size=32, value_type=OtherEntry, value_format=OtherFormat(timestamp="Q", result="B", extra="B")
        )
        table[H(1)] = OtherEntry(timestamp=123, result=1, extra=9)
        with io.BytesIO() as f:
            table.write(f)
            data = f.getvalue()
        repository.store_store(PackTracker.NAME, data + sha256(data).digest())

        tracker = PackTracker.load(repository.store)
        assert len(tracker) == 0


def test_check_progress_covers_packs_and_index(tmp_path, monkeypatch):
    # check() uses a separate progress indicator for index/ and for packs/. Each one is sized to its own
    # namespace and driven to 100% by a final show(current=total). A fake indicator records the wiring
    # without depending on log output.
    indicators = []

    class FakePI:
        def __init__(self, total=0, **kwargs):
            self.total = total
            self.position = 0
            indicators.append(self)

        def show(self, current=None, increase=0, *args, **kwargs):
            self.position = current if current is not None else self.position + increase

        def finish(self, *args, **kwargs):
            pass

    monkeypatch.setattr("borg.repository.ProgressIndicatorPercent", FakePI)
    pack = fchunk(b"A", chunk_id=H(1))
    pack_name = "packs/" + bin_to_hex(sha256(pack).digest())
    index_content = b"serialized chunk index"
    index_name = "index/" + bin_to_hex(sha256(index_content).digest())
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store(pack_name, pack)
        repository.store_store(index_name, index_content)
        # create() already wrote a chunk index, so don't assume a count: derive it from the store.
        n_packs = len(repository.store_list("packs"))
        n_index = len(repository.store_list("index"))
        assert repository.check(repair=False) is True
    # one indicator per namespace, each sized to its own object count ...
    assert sorted(pi.total for pi in indicators) == sorted([n_index, n_packs])
    # ... and each driven all the way to 100%.
    for pi in indicators:
        assert pi.position == pi.total


def test_pack_writer_final_partial_pack_uses_sha256():
    # A final flush with fewer pieces than max_count must still use SHA256(pack_bytes).
    store = MockStore()
    chunk_id = b"d" * 32
    cdata = b"solo"
    pw = PackWriter(store, max_count=3, chunks=ChunkIndex())
    assert pw.add(chunk_id, cdata) is None
    results = pw.flush()
    assert results is not None
    assert len(results) == 1
    _, pack_id, _, _ = results[0]
    assert pack_id == sha256(cdata).digest()
    assert pack_id != chunk_id


def test_pack_reader_in_memory_walks_objects():
    obj1 = fchunk(b"payload-one", meta=b"meta1", chunk_id=H(1))
    obj2 = fchunk(b"d2", meta=b"m2", chunk_id=H(2))
    pack = obj1 + obj2
    headers = PackReader(pack_contents=pack).iter_headers()
    assert list(headers) == [(H(1), 0, len(obj1)), (H(2), len(obj1), len(obj2))]


def test_pack_reader_empty_pack():
    assert list(PackReader(pack_contents=b"").iter_headers()) == []


def test_pack_reader_stops_on_trailing_partial_header():
    # a truncated trailing header is a clean stop, not an error
    obj = fchunk(b"data", meta=b"meta", chunk_id=H(3))
    pack = obj + b"\x00\x00\x00"
    assert list(PackReader(pack_contents=pack).iter_headers()) == [(H(3), 0, len(obj))]


def test_pack_reader_iter_headers_reads_through_store(tmp_path):
    obj1 = fchunk(b"FIRST", chunk_id=H(47))
    obj2 = fchunk(b"SECOND", chunk_id=H(48))
    pack = obj1 + obj2
    pack_id = H(43)
    with Repository(str(tmp_path / "repo"), exclusive=True, create=True) as repository:
        repository.store_store("packs/" + bin_to_hex(pack_id), pack)
        reader = PackReader(repository.store, pack_id)
        assert list(reader.iter_headers()) == [(H(47), 0, len(obj1)), (H(48), len(obj1), len(obj2))]



from cpython cimport PyMem_Malloc, PyMem_Free
from cpython.buffer cimport PyBUF_SIMPLE, PyObject_GetBuffer, PyBuffer_Release

cdef extern from "openssl/evp.h":
    ctypedef struct EVP_CIPHER:
        pass
    ctypedef struct EVP_CIPHER_CTX:
        pass
    ctypedef struct ENGINE:
        pass

    const EVP_CIPHER *EVP_aes_256_ctr()

    EVP_CIPHER_CTX *EVP_CIPHER_CTX_new()
    void EVP_CIPHER_CTX_free(EVP_CIPHER_CTX *a)

    int EVP_EncryptInit_ex(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher, ENGINE *impl,
                           const unsigned char *key, const unsigned char *iv)
    int EVP_DecryptInit_ex(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher, ENGINE *impl,
                           const unsigned char *key, const unsigned char *iv)
    int EVP_EncryptUpdate(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl,
                          const unsigned char *in_, int inl)
    int EVP_DecryptUpdate(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl,
                          const unsigned char *in_, int inl)
    int EVP_EncryptFinal_ex(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl)
    int EVP_DecryptFinal_ex(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl)


ctypedef const EVP_CIPHER * (* CIPHER)()


cdef Py_buffer ro_buffer(object data) except *:
    cdef Py_buffer view
    PyObject_GetBuffer(data, &view, PyBUF_SIMPLE)
    return view


cdef class AES:
    """A thin wrapper around the OpenSSL EVP cipher API - for legacy key file encryption."""
    cdef CIPHER cipher
    cdef EVP_CIPHER_CTX *ctx
    cdef unsigned char enc_key[32]
    cdef int cipher_blk_len
    cdef int iv_len
    cdef unsigned char iv[16]
    cdef long long blocks

    def __init__(self, enc_key, iv=None):
        assert isinstance(enc_key, bytes) and len(enc_key) == 32
        self.enc_key = enc_key
        self.iv_len = 16
        assert sizeof(self.iv) == self.iv_len
        self.cipher = EVP_aes_256_ctr
        self.cipher_blk_len = 16
        if iv is not None:
            self.set_iv(iv)
        else:
            self.blocks = -1  # make sure set_iv is called before encrypt

    def __cinit__(self, enc_key, iv=None):
        self.ctx = EVP_CIPHER_CTX_new()

    def __dealloc__(self):
        EVP_CIPHER_CTX_free(self.ctx)

    def encrypt(self, data, iv=None):
        if iv is not None:
            self.set_iv(iv)
        assert self.blocks == 0, 'iv needs to be set before encrypt is called'
        cdef Py_buffer idata
        cdef bint idata_acquired = False
        cdef unsigned char *odata = NULL
        cdef int ilen = len(data)
        cdef int olen = 0
        cdef int offset

        try:
            odata = <unsigned char *>PyMem_Malloc(ilen + self.cipher_blk_len)
            if not odata:
                raise MemoryError

            idata = ro_buffer(data)
            idata_acquired = True

            if not EVP_EncryptInit_ex(self.ctx, self.cipher(), NULL, self.enc_key, self.iv):
                raise Exception('EVP_EncryptInit_ex failed')
            offset = 0
            if not EVP_EncryptUpdate(self.ctx, odata, &olen, <const unsigned char*> idata.buf, ilen):
                raise Exception('EVP_EncryptUpdate failed')
            offset += olen
            if not EVP_EncryptFinal_ex(self.ctx, odata+offset, &olen):
                raise Exception('EVP_EncryptFinal failed')
            offset += olen
            self.blocks = self.block_count(offset)
            return odata[:offset]
        finally:
            if odata:
                PyMem_Free(odata)
            if idata_acquired:
                PyBuffer_Release(&idata)

    def decrypt(self, data):
        cdef Py_buffer idata
        cdef bint idata_acquired = False
        cdef unsigned char *odata = NULL
        cdef int ilen = len(data)
        cdef int offset
        cdef int olen = 0

        try:
            odata = <unsigned char *>PyMem_Malloc(ilen + self.cipher_blk_len)
            if not odata:
                raise MemoryError

            idata = ro_buffer(data)
            idata_acquired = True

            if not EVP_DecryptInit_ex(self.ctx, self.cipher(), NULL, self.enc_key, self.iv):
                raise Exception('EVP_DecryptInit_ex failed')
            offset = 0
            if not EVP_DecryptUpdate(self.ctx, odata, &olen, <const unsigned char*> idata.buf, ilen):
                raise Exception('EVP_DecryptUpdate failed')
            offset += olen
            if not EVP_DecryptFinal_ex(self.ctx, odata+offset, &olen):
                raise Exception('EVP_DecryptFinal failed')
            offset += olen
            self.blocks = self.block_count(ilen)
            return odata[:offset]
        finally:
            if odata:
                PyMem_Free(odata)
            if idata_acquired:
                PyBuffer_Release(&idata)

    def block_count(self, length):
        return (length + self.cipher_blk_len - 1) // self.cipher_blk_len

    def set_iv(self, iv):
        if isinstance(iv, int):
            iv = iv.to_bytes(self.iv_len, byteorder='big')
        assert isinstance(iv, bytes) and len(iv) == self.iv_len
        self.iv = iv
        self.blocks = 0

    def next_iv(self):
        iv = int.from_bytes(self.iv[:self.iv_len], byteorder='big')
        return iv + self.blocks
